"""RExam Item Editor

(c) Oliver Lindemann
"""

__version__ = '0.2.11'
__author__ = 'Oliver Lindemann'
APPNAME = "RExam Item Editor"
WEBSITE = "https://github.com/lindemann09/rexam-item-editor"

from sys import version_info as _vi
if _vi[0] != 3 or _vi[1] < 5:
    from .sysinfo import PYTHON_VERSION
    raise RuntimeError("{} {} ".format(APPNAME, __version__) +
                       "is not compatible with Python {0}".format(
                           PYTHON_VERSION) +
                       "\n\nPlease use Python 3.5 or higher.")




LANGUAGE1 = "Dutch"
LANGUAGE2 = "English"
CODE_L1 = "nl"
CODE_L2 = "en"

FILE_ENCODING = 'utf-8'

#GUI

COLOR_BKG_INACTIVE = "#8A8A8A"
COLOR_BKG_ACTIVE = "#FFFFFF"
COLOR_BKG_ACTIVE_INFO = "#f6f6f6"
COLOR_BKG_TAX_LIST = COLOR_BKG_ACTIVE_INFO
COLOR_QUEST = "#BBBBDD"
COLOR_SOLUTION = "#BBDDBB"
COLOR_RED_BTN = ("#000000", "#DDBBBB")
COLOR_GREEN_BTN = ("#000000", "#BBDDBB")

SG_COLOR_THEME = {'BACKGROUND': '#E0E0E2',
                  'TEXT': '#000000',
                  'INPUT': COLOR_BKG_ACTIVE_INFO,
                  'TEXT_INPUT': '#000000',
                  'SCROLL': '#99CC99',
                  'BUTTON': ('#000000', '#d4d7dd'),
                  'PROGRESS': ('#D1826B', '#CC8019'),
                  'BORDER': 1, 'SLIDER_DEPTH': 0,
                  'PROGRESS_DEPTH': 0, }

WIDTH_ML = 80 # multi line field for text input
LEN_ML_SMALL = 6
LEN_ML_LARGE = 15 # tab layout
LEN_ANSWER_SMALL = 5
LEN_ANSWER_LARGE = 8
TAB_LAYOUT = True

MAX_RECENT_DIRS = 5


from os import path
import PySimpleGUI as sg

from .. import sysinfo, __version__, __author__, APPNAME, templates
from ..consts import FILE_ENCODING
from ..rexam import r_render, extypes
from ..rexam.rmd_file import RmdFile, SEP, TAG_L1, TAG_L2, TAG_BILINGUAL
from ..rexam.item import RExamItem


def ask_save(item_name, txt=None):
    if txt is None:
        layout = []
    else:
        layout = [[sg.Text(txt)], [sg.Text("")]]
    layout.extend([[sg.Text("There are unsaved changes in '{}'.".format(
        item_name))],
              [sg.Save("Save", key="save"),
               sg.Cancel("Dismiss changes")]])
    window = sg.Window("{}".format(item_name), layout, finalize=True,
                       modal=True, keep_on_top=True)
    while True:
        window.refresh()
        event, v = window.read()
        break

    window.close()
    return event == "save"


def show_text_file(file, file2=None):
    win_titel = "View Files"
    content = [None, None]
    files = []
    for i, fl in enumerate((file, file2)):
        if isinstance(fl, RmdFile):
            win_titel = "View Files: {}".format(fl.name)
            fl = fl.full_path
        try:
            with open(fl, "r", encoding=FILE_ENCODING) as fl_hdl:
                content[i] = fl_hdl.readlines()
            files.append(fl)
        except:
            pass

    if len(files)==0:
        return

    tabs = []
    for c, f in zip(content, files):
        if c is not None:
            tabs.append(sg.Tab("{}".format(path.split(f)[1]),
                    [[sg.Text("File: {}".format(f))],
                    [sg.Multiline(default_text="".join(c),
                        disabled=True, size=(80, 40))]]))

    layout = [[sg.TabGroup([tabs])],
              [sg.CloseButton("Close")]]
    window = sg.Window(win_titel, layout, finalize=True)
    window.refresh()
    while True:
        window.refresh()
        event, v = window.read()
        break
    window.close()


def render(file):
    if isinstance(file, RmdFile):
        file = file.full_path
    r = r_render.RRender()
    if r.r_init_error is not None:
        sg.Print(r.r_init_error)
        return

    error = r.rmd_to_html(file)
    if error is None:
        r.open_html(new=0)
    else:
        sg.Print(error)


def about():
    old_theme = sg.theme()
    sg.theme("DarkBlack")
    width = 40

    layout = [[sg.Text(APPNAME, size=(width , 1), font='bold',
                       justification='center')],
              [sg.Text("Version {}".format(__version__), font='bold',
                       size=(width , 1),
                       justification='center')],
              [sg.Text(" " * 22), sg.Image(path.join(path.dirname(__file__),
                                                     "essb.png"))],
              [sg.Text("")] ]

    info_array = ["(c) {}".format(__author__), ""] + sysinfo.info()

    layout.append([sg.Multiline(default_text="\n".join(info_array),
                                size=(55, len(info_array)),
                                border_width=0,
                                background_color='black',
                                auto_size_text=True,
                                no_scrollbar=True,
                                text_color='white')])

    window = sg.Window("About", layout, finalize=True,
                       modal=True, keep_on_top=True)
    window.refresh()
    while True:
        window.refresh()
        event, v = window.read()
        break
    window.close()
    sg.theme(old_theme)


class FrameMakeName(object):

    def __init__(self, default_name=""):

        default_name = path.splitext(default_name)[0] # remove possible extension
        defaults = [""] * 3
        if default_name.endswith(TAG_BILINGUAL):
            defaults[-1] = "Bilingual"
            default_name = default_name[:-1*len(TAG_BILINGUAL)]
        elif default_name.endswith(TAG_L1):
            defaults[-1] = "Dutch"
            default_name = default_name[:-1*len(TAG_L1)]
        elif default_name.endswith(TAG_L2):
            defaults[-1] = "English"
            default_name = default_name[:-1*len(TAG_L2)]
        else:
            defaults[-1] = ""

        for c, txt in enumerate(default_name.split(SEP, maxsplit=2)):
            if c == len(defaults)-2: # cast number
                try:
                    defaults[c] = str(int(txt))
                except:
                    defaults[c-1] += SEP + txt
            else:
                defaults[c] = txt

        self.txt_name1 = sg.Text("", size=(43, 1), key="txt_name1")
        self.txt_name2 = sg.Text("", size=(43, 1), key="txt_name2")
        self.fr_names = sg.Frame("", [[sg.Text("a:", size=(2,1)),
                                  self.txt_name1],
                                 [sg.Text("b:" , size=(2,1)),
                                  self.txt_name2]],
                                  border_width=0)

        self.fln = [sg.InputText(default_text=defaults[0], size=(25,1),
                                 enable_events=True)] # could have potentiall
                                                     # multiple parts
        self.fln_cnt = sg.InputText(default_text=str(defaults[-1]), size=(4, 1),
                                    enable_events=True)

        self.fln_lang = sg.DropDown(default_value=defaults[-1],
                                    values=["Dutch", "English", "Bilingual"],
                                    enable_events=True)

        self.frame = sg.Frame("Item Name(s)",[
            [top_label([self.fln[0], sg.Text(SEP)], label="Topic"),
             top_label([self.fln_cnt, sg.Text(SEP)], label="Counter"),
             top_label(self.fln_lang, label="Language")],
             [self.fr_names]
             ])

    def update_names(self):
        # call me once after window created
        name_parts = map(lambda x: x.get().strip().lower(), self.fln)
        fln_cnt = self.fln_cnt.get().strip()

        name1 = SEP.join(filter(lambda x:len(x), name_parts))
        name2 = ""

        if len(name1):
            try:
                name1 += SEP + str(int(fln_cnt)).zfill(3)
            except:
                pass

        if len(name1) > 0:
            lang = self.fln_lang.get()
            if lang == "Dutch":
                name1 = name1 + TAG_L1
            elif lang == "English":
                name1 = name1 + TAG_L2
            elif lang == "Bilingual":
                name2 = name1 + TAG_L2
                name1 = name1 + TAG_L1

        self.txt_name1.update(value=name1)
        self.txt_name2.update(value=name2)

    @property
    def name1(self):
        return self.txt_name1.get()

    @property
    def name2(self):
        return self.txt_name2.get()


def new_item(base_directory):
    fr_make_name = FrameMakeName()
    lb_type = sg.Listbox(values=["None"] + list(extypes.EXTYPES.values()),
                         default_values=["None"],
                         select_mode=sg.LISTBOX_SELECT_MODE_SINGLE,
                         size=(26, 6), no_scrollbar=True,
                         key="lb_type")

    layout = [[fr_make_name.frame]]
    layout.append([sg.Frame("Template", [[lb_type]]),
                   sg.Cancel(size=(10, 2)),
                   sg.Button("Create", size=(10, 2), key="create")])
    window = sg.Window("New Item(s)", layout, finalize=True,
                       modal=True, keep_on_top=True)
    fr_make_name.update_names()
    while True:
        window.refresh()
        event, v = window.read()
        if event in ("Cancel", "create", None):
            break
        else:
            fr_make_name.update_names()

    item1, item2 = (None, None)
    if event == "create":
        # template
        sel = lb_type.get_indexes()[0]
        if sel > 0:
            template_key = list(extypes.EXTYPES.keys())[sel - 1]
        else:
            template_key = None

        if len(fr_make_name.name1):
            rel_path = path.join(fr_make_name.name1,
                            "{}{}".format(fr_make_name.name1,RmdFile.SUFFIX))
            item1 = RExamItem(RmdFile(file_path=rel_path,
                                    base_directory = base_directory))
            if template_key is not None:
                item1.import_file(templates.FILES[template_key])
        if len(fr_make_name.name2):
            rel_path = path.join(fr_make_name.name2,
                        "{}{}".format(fr_make_name.name2, RmdFile.SUFFIX))
            item2 = RExamItem(RmdFile(file_path=rel_path,
                                    base_directory = base_directory))
            if template_key is not None:
                item2.import_file(templates.FILES[template_key])

    window.close()

    return item1, item2

def rename_item(item_name):
    fr_make_name = FrameMakeName(item_name)
    fix_dir = sg.Checkbox(text="Adapt directory name", default=True)
    layout = [[fr_make_name.frame]]
    layout.append([sg.Frame("", [[fix_dir]]),
                   sg.Text(" "*10),
                   sg.Cancel(size=(10, 2)),
                   sg.Button("Rename", size=(10, 2), key="rename")])
    window = sg.Window("Rename", layout, finalize=True,
                       modal=True, keep_on_top=True)
    fr_make_name.update_names()

    while True:
        window.refresh()
        event, _ = window.read()
        if event in ("Cancel", "rename", None):
            break
        else:
            fr_make_name.update_names()

    window.close()
    if event=="rename":
        return fr_make_name.name1, fr_make_name.name2, bool(fix_dir.get())
    else:
        return None, None, None

#TODO rendering picture
#TODO rename no language


from os import path, rename
from collections import OrderedDict
import hashlib
from copy import deepcopy
import textwrap

from . import extypes
from .. import templates
from ..consts import FILE_ENCODING
from .rmd_file import RmdFile
from .filepath import FilePath
from .issue import Issue
from ..misc import extract_parameter


class ItemSection(object):

    def __init__(self, parent, label, underline_chr, min_underline_length=4):

        assert (isinstance(parent, (ItemSection, RExamItem)))

        self._parent = parent
        self._underline_chr = underline_chr
        self._underline_string = underline_chr * min_underline_length

        self.label = label
        self.text_array = []  # array of text lines ending with \n (like readlines)
        self.line_range = [None, None]
        self.answer_list = None

    def parse(self, parent=None):
        # find the section and copy content to object
        if parent is not None:
            self._parent = parent
        prev = ""
        is_section = False
        self.text_array = []
        self.line_range = [None, None]
        cnt = None
        for cnt, line in enumerate(self._parent.text_array):
            if not is_section:
                if line.startswith(self._underline_string) and \
                        prev.strip() == self.label:
                    is_section = True
                    self.line_range[0] = cnt - 1
            else:
                if line.startswith(self._underline_string):
                    # end section
                    self.text_array.pop()
                    cnt -= 2
                    break
                else:
                    self.text_array.append(line)
            prev = line

        if self.line_range[0] is not None:
            self.line_range[1] = cnt

        if not isinstance(self, AnswerList):
            answer_list = AnswerList(self)
            answer_list.parse()
            a, b = answer_list.line_range
            if a is not None and b is not None:
                self.text_array = self.text_array[:a] + self.text_array[b + 1:]
                self.answer_list = answer_list
            else:
                self.answer_list = None

    def add_answer_list_section(self):
        if not self.has_answer_list_section():
            self.answer_list = AnswerList(self)

    def has_answer_list_section(self):
        return self.answer_list is not None

    def str_markdown_heading(self):
        return "{}\n{}\n".format(self.label,
                                 str(self._underline_chr * len(self.label)))

    def str_text(self, ignore_empty_lines=False, wrap_text_width=0):
        if ignore_empty_lines:
            tmp = [x for x in self.text_array if len(x.strip())]
        else:
            tmp = self.text_array

        if wrap_text_width > 1:
            rtn = ""
            for x in tmp:
                rtn += textwrap.fill(x, width=wrap_text_width,
                                     replace_whitespace=False) + "\n"
        else:
            rtn = "".join(tmp).rstrip()

        return rtn.rstrip()

    def str_text_short(self, max_lines=2, ignore_empty_lines=True):
        """return x lines for the section and ignores empty lines"""
        if ignore_empty_lines:
            cnt = 0
            rtn = ""
            for x in self.text_array:
                if len(x.strip()):
                    rtn += x
                    cnt += 1
                    if cnt >= max_lines:
                        break
            return rtn.strip()

        else:
            return "".join(self.text_array[:max_lines]).strip()

    def __str__(self):
        # section as string
        rtn = self.str_markdown_heading() + self.str_text()
        if self.answer_list is not None:
            return rtn + "\n\n" + str(self.answer_list)
        else:
            return rtn


class AnswerList(ItemSection):
    TAG_CORRECT = "#"
    TAG_ITEM = "*"

    def __init__(self, parent):
        super().__init__(parent, "Answerlist", "-")
        self.answers = []
        self._correct = []
        self._tab_sep = []  # used tab separator for this item?

    def parse(self, parent=None):
        super().parse(parent=parent)
        self.answers = []
        self._correct = []
        self._tab_sep = []
        unparsed_content = []

        l_tag_item = len(AnswerList.TAG_ITEM) + 1
        l_tag_corr = len(AnswerList.TAG_CORRECT) + 1
        while len(self.text_array) > 0:
            answer = self.text_array.pop(0)
            tag, tab_sep = check_tag(answer, AnswerList.TAG_ITEM, AnswerList.TAG_CORRECT)
            if tab_sep is not None:
                self._tab_sep.append(tab_sep)
            if tag == AnswerList.TAG_ITEM:
                self.answers.append(answer[l_tag_item:].strip())
                self._correct.append(False)
            elif tag == AnswerList.TAG_CORRECT:
                self.answers.append(answer[l_tag_corr:].strip())
                self._correct.append(True)
            else:
                unparsed_content.append(answer)

        self.text_array = unparsed_content

    @staticmethod
    def extract_solution(markdown):
        """extracts solution from markdown string that used TAG_CORRECT"""

        solution = ""
        for l in markdown.split("\n"):
            tag, _ = check_tag(l, AnswerList.TAG_ITEM, AnswerList.TAG_CORRECT)
            if tag == AnswerList.TAG_ITEM:
                solution += "0"
            elif tag == AnswerList.TAG_CORRECT:
                solution += "1"
        return solution

    @property
    def solution_str(self):
        return "".join(map(lambda x: str(int(x)), self._correct))

    @solution_str.setter
    def solution_str(self, value):
        self._correct = []
        for cnt, ans in enumerate(self.answers):
            try:
                self._correct.append(value[cnt] == "1")
            except:
                self._correct.append(False)

    def str_answers(self, tag_mark_correct=False, highlight_correct_before_after=""):
        """mark_correct add chars before and after correct e.g. '**' for markdown bold"""
        rtn = ""
        for ans, correct, tab_sep in zip(self.answers, self._correct, self._tab_sep):
            if tag_mark_correct and correct:
                tag = AnswerList.TAG_CORRECT
            else:
                tag = AnswerList.TAG_ITEM

            if correct:
                mark = highlight_correct_before_after
            else:
                mark = ""

            if tab_sep:
                rtn += "{}\t{}{}{}\n".format(tag, mark, ans, mark)
            else:
                rtn += "{} {}{}{}\n".format(tag, mark, ans, mark)

        return rtn

    def __str__(self):
        rtn = self.str_markdown_heading() + self.str_answers() + self.str_text()
        return rtn.strip()

    def fix_tabs(self):
        if True in self._tab_sep:
            self._tab_sep = [False] * len(self._tab_sep)

    def validate(self):
        issues = []
        if True in self._tab_sep:
            issues.append(Issue("tabs_in_answerlist",
                                "Tabs used as separator in Answerlist",
                                self.fix_tabs))
        return issues


class ItemMetaInfo(ItemSection):
    _REQUIRED_PARAMETER = None

    def __init__(self, parent):
        assert (isinstance(parent, RExamItem))
        super().__init__(parent, "Meta-information", "=")
        self.parameter = OrderedDict()

    @classmethod
    def required_parameter(cls):
        if cls._REQUIRED_PARAMETER is None:
            # try to load it once from templates
            rtn = {}
            for k, filename in templates.FILES.items():
                meta_info = False
                rtn[k] = OrderedDict()
                with open(filename, "r", encoding=FILE_ENCODING) as fl:
                    for l in fl:
                        if l.startswith("Meta-information"):
                            meta_info = True
                        elif meta_info:
                            para = extract_parameter(l)
                            if para is not None:
                                rtn[k].update(para)

            cls._REQUIRED_PARAMETER = rtn

        return cls._REQUIRED_PARAMETER

    def parse(self, parent=None, reset_parameter=False):
        super().parse(parent=parent)
        additional_content = []
        if reset_parameter:
            self.parameter = OrderedDict()
        while len(self.text_array) > 0:
            l = self.text_array.pop(0)
            para = extract_parameter(l)
            if para is None:
                additional_content.append(l)
            else:
                self.parameter.update(para)
        self.text_array = additional_content

    def str_parameter(self):
        rtn = ""
        for k, v in self.parameter.items():
            rtn += "{}: {}\n".format(k, v)
        return rtn

    def __str__(self):
        rtn = self.str_markdown_heading() + self.str_parameter() + "\n" \
              + self.str_text()
        return rtn.strip()

    def _try_parameter(self, label):
        try:
            return self.parameter[label]
        except:
            return ""

    def sort_parameter(self):
        # sorting parameter based on Templates (in place)
        required = ItemMetaInfo.required_parameter()
        try:
            req_para = required[self.type].keys()
        except:
            return False

        new_para = OrderedDict()
        for k in req_para:
            if k in self.parameter:
                new_para[k] = self.parameter.pop(k)
        new_para.update(self.parameter)
        self.parameter = new_para

    @property
    def name(self):
        return self._try_parameter("exname")

    @name.setter
    def name(self, v):
        self.parameter["exname"] = v

    @property
    def section(self):
        return self._try_parameter("exsection")

    @section.setter
    def section(self, v):
        self.parameter["exsection"] = v

    @property
    def type(self):
        return self._try_parameter("extype")

    @type.setter
    def type(self, v):
        self.parameter["extype"] = v

    @property
    def solution(self):
        return self._try_parameter("exsolution")

    @solution.setter
    def solution(self, v):
        self.parameter["exsolution"] = v

    def check_type(self):
        """:returns True if known type
        """

        return self.type in extypes.EXTYPES.keys()

    def requires_answer_list(self):
        return self.type in extypes.HAVE_ANSWER_LIST

    def get_missing_parameter(self):
        required = ItemMetaInfo.required_parameter()
        try:
            parameter = required[self.type]
        except:
            return {}

        return {k: v for k, v in parameter.items() if k not in self.parameter}

    def fix_missing_parameter(self):
        for k, v in self.get_missing_parameter().items():
            self.parameter[k] = v

    def validate(self):
        issues = []

        # is type defined type
        if not self.check_type():
            issues.append(Issue("extype",
                                "Unknown/undefined item type(extype))", None))

        # missing parameter
        missing_parameter = self.get_missing_parameter()
        if len(missing_parameter):
            issues.append(Issue("missings", "Missing required meta-information: {}".format(
                list(missing_parameter.keys())),
                                self.fix_missing_parameter))

        return issues


class RExamItem(RmdFile):
    META_INFO_CLASS = ItemMetaInfo

    def __init__(self, file_path=None):
        assert isinstance(file_path, FilePath) or file_path is None
        if file_path is not None:
            super().__init__(file_path=file_path.relative_path,
                             base_directory=file_path.base_directory)
        else:
            super().__init__(file_path)
        self.question = ItemSection(self, "Question", "=")
        self.solution = ItemSection(self, "Solution", "=")
        self.meta_info = RExamItem.META_INFO_CLASS(self)

        self.header = []
        self.text_array = []
        self._hash = None

        if path.isfile(self.full_path):
            self.import_file(self.full_path)

    def hash(self):
        """question id is based on the file content"""

        if self._hash is None:
            txt = "".join(self.text_array)
            self._hash = hashlib.sha1(txt.encode()).hexdigest()  # calc only once

        return self._hash

    def hash_short(self):
        return self.hash()[:7]

    def import_file(self, text_file):
        """import a text file as content"""
        self.header = []
        self.text_array = []
        with open(text_file, "r", encoding=FILE_ENCODING) as fl:
            self.parse(fl.readlines())

    def parse(self, source_text, reset_meta_information=False):
        """parse file or source text is specified"""
        if isinstance(source_text, str):
            self.text_array = list(map(lambda x: x + "\n",
                                       source_text.split("\n")))
            # array of text lines ending with \n (like readlines)
        else:
            self.text_array = source_text

        self.question.parse(parent=self)
        self.solution.parse(parent=self)
        self.meta_info.parse(parent=self,
                             reset_parameter=reset_meta_information)

        self.header = deepcopy(self.text_array[:self.question.line_range[0]])
        # override answer_list correctness with meta info solution
        self.update_solution(self.meta_info.solution)

    def __str__(self):
        rtn = "".join(self.header)
        rtn += str(self.question) + "\n\n\n" + \
               str(self.solution) + "\n\n\n" + str(self.meta_info)
        return rtn

    def save(self):
        if len(self.full_path):
            self.make_dirs()
            with open(self.full_path, "w", encoding=FILE_ENCODING) as fl:
                fl.write(str(self))

    def fix_add_answer_list(self):
        self.question.add_answer_list_section()

    def fix_directory_name(self):
        self.save()
        rename(self.directory, path.join(self.base_directory, self.name))

    def fix_uppercases_in_relative_path(self):
        self.save()
        old = self.full_path
        self.name = self.name.lower()
        rename(old, self.full_path)
        self.fix_directory_name()

    def validate(self):
        """Validates the item and returns a list of issues"""

        issues = self.meta_info.validate()
        if self.question.has_answer_list_section():
            issues.extend(self.question.answer_list.validate())

        # check answer & feedback list
        if self.meta_info.requires_answer_list():
            if not self.question.has_answer_list_section():
                issues.append(Issue("answers", "No answer list defined",
                                    self.fix_add_answer_list))
        else:
            if self.question.has_answer_list_section():
                issues.append(Issue("answers", "Answer list not required"))

        return issues

    def update_solution(self, solution_str):
        if len(solution_str) == 0 and len(self.meta_info.solution) == 0:
            # don't write solution nothing changed (avoid creating of solution
            # is solution_str is empty and parameter is not yet defined)
            return
        self.meta_info.solution = solution_str
        self.meta_info.sort_parameter()
        if self.question.has_answer_list_section():
            self.question.answer_list.solution_str = solution_str

    @staticmethod
    def load(file_path, base_directory):
        if file_path is None:
            return None
        else:
            return RExamItem(RmdFile(file_path,
                                     base_directory=base_directory))

    def markdown(self, enumerator=None, wrap_text_width=0, tag_mark_correct=True, highlight_correct_char=""):
        '''optional counter to enumerate questions'''
        question_str = self.question.str_text(ignore_empty_lines=False,
                                              wrap_text_width=wrap_text_width)

        if self.question.answer_list is not None:
            question_str += "\n\n" + \
                            self.question.answer_list.str_answers(
                                tag_mark_correct=tag_mark_correct,
                                highlight_correct_before_after=highlight_correct_char)

        if len(question_str.strip()):
            rtn = "# Question "
            if enumerator is not None:
                rtn += "{} ".format(enumerator)
            rtn += "({})\n\n{}".format(self.name, question_str)
            return rtn
        else:
            return ""


def check_tag(textline, tag1, tag2):
    """ helper function
    return (found tag[or None], is tab-separated)
    """
    l = textline.strip()
    if l.startswith(tag1 + " "):
        return tag1, False
    elif l.startswith(tag2 + " "):
        return tag2, False
    if l.startswith(tag1 + "\t"):
        return tag1, True  # item, using tab
    elif l.startswith(tag2 + "\t"):
        return tag2, True  # correct, using tab
    else:
        return None, None



import os
import tempfile
import re


def replace_list_element(lst, source_idx, target_idx):
    """replaces an element in a list"""
    if source_idx < len(lst) and target_idx<len(
    lst):
        tmp = lst.pop(source_idx)
        return lst[:target_idx] + [tmp] + lst[target_idx:]
    else:
        return []

def subdict(d, nested_keys=None):
    """:return the dict nested hierarchically indicated by nested_keys
    or None if key list is incorrect
    :param nested_keys list of keys or a single keys

    """
    if not isinstance(nested_keys, (tuple, list)):
        nested_keys = [nested_keys]
    for k in nested_keys:
        try:
            d = d[k]
        except:
            return {}
    return d

def splitstrip(text, sep):
    return list(map(lambda x: x.strip(), text.split(sep)))

def yesno(bool):
    if bool:
        return "Yes"
    else:
        return "No"

def get_temp_dir(appname, make_dir=True):
    # creates and returns a temp folder

    tmpdir = tempfile.gettempdir()
    tmpdir = os.path.join(tmpdir, appname)
    if make_dir:
        try:
            os.mkdir(tmpdir)
        except:
            pass

    return tmpdir

class CaseInsensitiveStringList(object):
    """String list that handles string search case insensitive"""

    def __init__(self, str_list=()):
        self._str_list = list(str_list)
        self._str_lower = [x.lower() for x in self._str_list]

    def __len__(self):
        return len(self._str_list)

    def append(self, new_string):
        self._str_list.append(new_string)
        self._str_lower.append(new_string.lower())

    def pop(self, index=-1):
        self._str_lower.pop(index)
        return self._str_list.pop(index)

    def remove(self, element):
        """removes element and returns it, raises exception in not included"""
        element = str(element).lower()
        idx = self._str_lower.index(element)
        self._str_lower.pop(idx)
        return self._str_list.pop(idx)

    def remove_all(self, element):
        element = str(element).lower()
        while True:
            try:
                idx = self._str_lower.index(element)
            except:
                break

            self._str_list.pop(idx)
            self._str_lower.pop(idx)

    def __contains__(self, item):
        return str(item).lower() in self._str_lower

    def get(self):
        return self._str_list

def remove_all(str_list, element, ignore_cases=False):
    """removes all occurrences of element from string list and ignores
    optionally letter cases"""

    if ignore_cases:
        return [e for e in str_list \
                    if str(e).lower() != str(element).lower()]
    else:
        return [e for e in str_list if e != element]

def extract_parameter(txt):
    # extract parameter for text line

    m = re.match(r"\s*\w+[\[\]\w]+:", txt)
    if m is not None:
        return {txt[:m.end()-1].strip(): txt[m.end():].strip()}
    return None

def iter_list(data):
    """Generates iterator over the data.
    If None, iterator over empty list. If data is not a list or a tuple,
    iterator over list with one one element [data]
    """
    if data is None:
        return iter([])
    elif isinstance(data, (list, tuple)):
        return iter(data)
    else:
        return iter([data])




from os import path, getcwd
import PySimpleGUI as sg

from .. import __version__, APPNAME, consts
from ..rexam.rmd_file import RmdFile
from ..rexam.item_database import BiLingRmdFileList
from ..rexam.r_render import RPY2INSTALLED
from ..rexam.item import RExamItem, AnswerList
from . import dialogs
from .json_settings import JSONSettings
from .gui_misc import GUIItem, GUIBaseDirectory, set_font
from .log import log


LANG1_EVENT_PREFIX = "Lang1"
LANG2_EVENT_PREFIX = "Lang2"


class MainWin(object):

    def __init__(self, clear_settings=False, change_meta_info_button=False,
                 monolingual=False, font=None, font_size=9):

        """languages is none or a list of two strings indicating the
        respective languages languages"""

        self.settings = JSONSettings(
                         appname=APPNAME.replace(" ", "_").lower(),
                         settings_file_name="settings.json",
                         defaults= {"recent_dirs": [], "bilingual": True},
                         reset=clear_settings)

        self.is_bilingual = not monolingual and self.settings.bilingual

        # remove not existing recent dirs
        existing_dirs =  list(filter(path.isdir, self.settings.recent_dirs))
        if len(existing_dirs) != len(self.settings.recent_dirs):
            self.settings.recent_dirs = existing_dirs
            self.settings.save()

        # LAYOUT
        set_font(font, font_size)
        if self.is_bilingual:
            two_languages = (consts.LANGUAGE1, consts.LANGUAGE2)
        else:
            two_languages = ("File", "")
        self.ig_l1 = GUIItem(two_languages[0], LANG1_EVENT_PREFIX,
                             show_hash=False,
                             change_meta_info_button=change_meta_info_button)
        self.ig_l2 = GUIItem(two_languages[1], LANG2_EVENT_PREFIX,
                             show_hash=self.ig_l1.show_hash,
                             change_meta_info_button=change_meta_info_button,
                             disabled=not self.is_bilingual)

        self.lb_items = sg.Listbox(values=[], enable_events=True,
                                   key="lb_files", size=(30, 39))

        self.txt_item_cnt = sg.Text("...", size=(28, 1))
        self.fr_items = sg.Frame("Items", [
            [self.txt_item_cnt],
            [self.lb_items],
            [sg.Button(button_text="<<", size=(11, 1), key="btn_previous"),
             sg.Button(button_text=">>", size=(11, 1), key="btn_next")]])

        self.btn_save = sg.Button(button_text="Save", size=(28, 2),
                                  disabled=True,
                                  key="save")

        self.gui_base_directory = GUIBaseDirectory(database_folder=self.base_directory,
                                                   key="btn_change_dir")

        self.txt_name = sg.Text("", size=(30, 1),
                                background_color=consts.COLOR_BKG_ACTIVE_INFO)
        self.btn_rename = sg.Button(button_text="Rename", size=(10, 1),
                                    key="rename")
        self.btn_second_lang = sg.Button(button_text="Add Language",
                                         size=(10, 1), disabled=True,
                                    key="second_lang")
        fr_item_name = sg.Frame("Item Name",
                             [[self.txt_name,
                              self.btn_rename,
                               self.btn_second_lang]])

        fr_btns =sg.Frame("", [[self.btn_save]])
        left_frame = sg.Frame("", [[self.fr_items], [fr_btns]],
                              border_width=0)

        self.menu = sg.Menu(menu_definition=self.menu_definition(),
                            tearoff=False)

        if self.is_bilingual:
            right_frames = [self.ig_l1.gui_element, self.ig_l2.gui_element]
        else:
            right_frames = [self.ig_l1.gui_element]

        self.layout = [
                  [self.menu],
                  [self.gui_base_directory.frame, fr_item_name],
                  [left_frame] + right_frames]

        self.fl_list = BiLingRmdFileList(files_first_level=True, # TODO somewhere in seetings
                                         files_second_level=True,
                                         check_for_bilingual_files=self.is_bilingual)

        self._unsaved_item = None

    def menu_definition(self):
        file = ['&New Item', '&Save Item', '---',
                'Open &Directory',
                'Recent', list(reversed(self.settings.recent_dirs[:-1])),
                '---','&Reload Item List',
                "---",  "Run &Exam Compiler",
                'C&lose']

        view = ["&Raw files", "---"]
        if not self.ig_l1.show_hash:
            view += ["Show &hashes"]
        else:
            view += ["Hide &hashes"]

        if self.is_bilingual:
            view += ["Single &Language Mode"]
        else:
            view += ["Bi&lingual Mode"]
        view += ['---', '&About']

        menu = [['&File', file], ["&View", view]]

        if  RPY2INSTALLED:
            d_inactive, e_inactive = "", ""
            if not self.ig_l1.is_activated():
                d_inactive="!"
            if not self.ig_l2.is_activated():
                e_inactive = "!"
            rmenu = ["&Render", ["{}&Dutch Version::render".format(d_inactive),
                            "{}&English Version::render".format(e_inactive)]]
            menu = menu[0], rmenu, menu[1]

        return menu

    @property
    def base_directory(self):
        try:
            return self.settings.recent_dirs[-1]
        except:
            return ""

    @base_directory.setter
    def base_directory(self, v):
        # update recent_dir list
        if len(self.settings.recent_dirs)>0 and v == self.settings.recent_dirs[-1]:
            return # nothing changes

        while True:
            try:
                self.settings.recent_dirs.remove(v)
            except:
                break
        self.settings.recent_dirs.append(v)
        self.settings.recent_dirs = self.settings.recent_dirs[
                                    -1 * consts.MAX_RECENT_DIRS:]  # limit n elements  self.update_item_list()

    @property
    def idx_selected_item(self):
        try:
            return self.lb_items.get_indexes()[0]
        except:
            return None

    @idx_selected_item.setter
    def idx_selected_item(self, index):
        n = len(self.lb_items.get_list_values())
        if n<=0:
            return
        elif index<0:
           index = 0
        elif index > n-1:
            index = n-1
        self.lb_items.update(set_to_index=index)
        self.load_selected_item()

    def select_item_by_filename(self, item_name):
        """select by file name"""

        idx = self.fl_list.find_filename(item_name)
        if idx is not None:
            self.idx_selected_item = idx

        return idx

    @property
    def unsaved_item(self):
        return self._unsaved_item

    @unsaved_item.setter
    def unsaved_item(self, v):
        self._unsaved_item = v
        self.btn_save.update(disabled=v is None)

    def update_name(self):
        if self.idx_selected_item is not None:
            #update name
            names = self.fl_list.get_shared_names(bilingual_tag=False)
            self.txt_name.update(value=names[self.idx_selected_item])
            self.btn_rename.update(disabled=False)
            if self.fl_list.files[self.idx_selected_item].is_bilingual():
                self.btn_second_lang.update(disabled=True)
            else:
                self.btn_second_lang.update(disabled=False)
        else:
            self.txt_name.update(value="")
            self.btn_rename.update(disabled=True)
            self.btn_second_lang.update(disabled=True)

        # dir information
        self.gui_base_directory.update_folder(self.base_directory)

    def update_item_list(self):
        if not path.isdir(self.base_directory):
            self.base_directory = sg.PopupGetFolder("Please select item directory:",
                title="{} ({})".format(APPNAME, __version__))
            if not path.isdir(self.base_directory):
                sg.PopupError("No valid item directory selected.")
                exit()

        self.fl_list = BiLingRmdFileList(base_directory=self.base_directory,
                                         files_first_level=self.fl_list.files_first_level,
                                         files_second_level=self.fl_list.files_second_level,
                                         check_for_bilingual_files=self.is_bilingual)

        cnt = self.fl_list.get_count()
        self.fr_items.update(value="{} items".format(cnt["total"]))

        cnt_txt = "{} {}, {} {}, {} {}/{}".format(
             cnt[consts.CODE_L1], consts.CODE_L1,
             cnt[consts.CODE_L2], consts.CODE_L2,
             cnt["bilingual"], consts.CODE_L1, consts.CODE_L2)
        if cnt["undef"] > 0:
            cnt_txt += ", {} undef".format(cnt["undef"])
        self.txt_item_cnt.update(value=cnt_txt)

        list_display = self.fl_list.get_shared_names()
        self.lb_items.update(values=list_display)

    def reset_gui(self, select_by_filename=None):

        if select_by_filename:
            selected_file = select_by_filename
        else:
            try:
                selected_file = self.fl_list.files[
                                    self.idx_selected_item].rmdfile_l1.filename
            except:
                selected_file = None

            self.update_item_list()
        self.ig_l1.rexam_item = None
        self.ig_l2.rexam_item = None
        self.ig_l2.update_gui()
        self.ig_l1.update_gui()
        self.menu.update(menu_definition=self.menu_definition())
        self.update_name()
        self.unsaved_item = None

        if selected_file is not None:
            self.select_item_by_filename(selected_file)

    def run(self):
        win = sg.Window("{} ({})".format(APPNAME, __version__),
                        self.layout, finalize=True, return_keyboard_events=True,
                        enable_close_attempted_event=True)

        if len(self.settings.recent_dirs) == 0: # very first launch
            self.base_directory = getcwd()

        self.reset_gui()
        self.idx_selected_item = 0

        while True:
            win.refresh()
            event, values = win.read(timeout=5000)

            if event == sg.WINDOW_CLOSE_ATTEMPTED_EVENT or \
                    event == "Close" or event is None:
                self.save_items(ask=True)
                restart = False
                break

            elif event.startswith(LANG1_EVENT_PREFIX) or event.startswith(LANG2_EVENT_PREFIX):
                #ItemGUI event
                self.process_item_gui_event(event, values)

                if self.unsaved_item is None: # if change in any text boxes
                    self.unsaved_item = self.idx_selected_item

            restart = self.process_event(event, values)
            if restart:
                break

        win.close()
        self.settings.save()

        if restart == "Run Exam Compiler":
            from .exam_compiler import ExamCompiler
            return ExamCompiler().run(database_folder=self.base_directory)
        elif restart:
            return MainWin().run()

    def process_item_gui_event(self, event, values):
        # ItemGUI events have to start with LANG1_PREFIX or LANG2_PREFIX
        is_l1_event = event.startswith(LANG1_EVENT_PREFIX)
        if not self.is_bilingual or is_l1_event:
            ig = self.ig_l1
        else:
            ig = self.ig_l2

        if event.endswith("dd_types"):
            ig.update_item()
            ig.rexam_item.meta_info.type = values[event]
            ig.rexam_item.meta_info.sort_parameter()
            ig.update_gui()

        elif event.endswith("btn_add_answer_list"):
            ig.rexam_item.question.add_answer_list_section()
            ig.update_gui()

        elif event.endswith("btn_add_feedback_list"):
            ig.rexam_item.solution.add_answer_list_section()
            ig.update_gui()

        elif event.endswith("_answer"):
            ig.update_answer_list_button()

        elif event.endswith("btn_update_exsolution"):
            ig.update_item()
            ig.rexam_item.update_solution(solution_str= \
                AnswerList.extract_solution(
                    ig.ml_answer.get()))
            ig.update_gui()

        elif event.endswith("btn_fix_meta_issues"):
            ig.update_item()
            for i in ig.rexam_item.validate():
                i.fix()
                if i.fix_requires_gui_reset:
                    if isinstance(i.fix_requires_gui_reset, str):
                        self.reset_gui(select_by_filename=i.fix_requires_gui_reset)
                    else:
                        self.reset_gui()
                    return
            ig.update_gui()

    def process_event(self, event, _values):
        """returns None or event if a restart is required"""

        if event == "__TIMEOUT__":
            if self.fl_list.is_file_list_changed():
                self.save_items(ask=True, info_text=
                        "Changes in base directory detected.")
                self.reset_gui()

        elif event in ("lb_files", "btn_next", "btn_previous"):
            self.save_items(ask=True)
            if event == "btn_next":
                if self.idx_selected_item is None:
                    self.idx_selected_item = 0
                else:
                    self.idx_selected_item += 1
            elif event == "btn_previous":
                if self.idx_selected_item is None:
                    self.idx_selected_item = 0
                else:
                    self.idx_selected_item -= 1

            else: # selected
                self.load_selected_item()

        elif event in ["Open Directory", "btn_change_dir"] or event in self.settings.recent_dirs:
            self.save_items(ask=True)
            if event in ["Open Directory", "btn_change_dir"]:
                fld = sg.PopupGetFolder(message="", no_window=True)
                if len(fld):
                    self.base_directory = fld
            else:
                self.base_directory = event  # from recent dir submenu
            self.reset_gui()

        elif event == "save" or event == "Save Item":
            self.save_items(ask=False)

        elif event == "New Item":
            self.new_item()

        elif event == "About":
            dialogs.about()

        elif event == "Reload Item List":
            self.reset_gui()

        elif event == "rename":
            self.rename()

        elif event == "second_lang":
            self.save_items()
            self.add_second_language()

        elif event == "Raw files":
            try:
                flns = self.fl_list.files[self.idx_selected_item]
            except:
                return

            self.save_items(ask=True)
            dialogs.show_text_file(flns.rmdfile_l1, flns.rmdfile_l2)

        elif event == "Show hashes" or event == "Hide hashes": # toggle hashes
            self.ig_l1.show_hash = not self.ig_l1.show_hash
            self.ig_l2.show_hash = self.ig_l1.show_hash
            self.reset_gui()

        elif event == "Single Language Mode" or event == "Bilingual Mode":
            self.settings.bilingual = event == "Bilingual Mode"
            self.settings.save()
            self.save_items(ask=True)
            return event

        elif event == "Run Exam Compiler":
            self.save_items(ask=True)
            return event

        elif event.endswith("render"):
            try:
                flns = self.fl_list.files[self.idx_selected_item]
            except:
                return
            if event.startswith("Dutch"):
                fl = flns.rmdfile_l1
            else:
                fl = flns.rmdfile_l2
            if fl is not None:
                self.save_items(ask=True)
                dialogs.render(fl)

    def load_selected_item(self):
        if self.idx_selected_item is None:
            return

        try:
            fls = self.fl_list.files[self.idx_selected_item]
        except:
            fls = None

        if fls is not None:
            if self.is_bilingual and fls.rmdfile_l1.language_code == consts.CODE_L2 \
                    and not fls.is_bilingual():
                self.ig_l2.rexam_item = RExamItem.load(fls.rmdfile_l1.full_path,
                                                       base_directory = self.base_directory)
                self.ig_l1.rexam_item = None
            else:
                self.ig_l1.rexam_item = RExamItem.load(fls.rmdfile_l1.full_path,
                                                       base_directory = self.base_directory)

                if fls.rmdfile_l2 is not None:
                    self.ig_l2.rexam_item = RExamItem.load(
                                    fls.rmdfile_l2.full_path,
                                    base_directory=self.base_directory)
                else:
                    self.ig_l2.rexam_item = None

        self.ig_l1.update_gui()
        self.ig_l2.update_gui()
        self.update_name()
        self.menu.update(menu_definition=self.menu_definition())

    def save_items(self, ask=False, info_text=None):
        if self.unsaved_item is not None:
            if ask:
                item_name = self.lb_items.get_list_values()[self.unsaved_item]
                if not dialogs.ask_save(item_name, txt = info_text):
                    self.unsaved_item = None
                    return
            self.ig_l1.save_item()
            self.ig_l2.save_item()
            self.unsaved_item = None

    def new_item(self, new_rmd_file_name = None):
        if new_rmd_file_name is None:
            new_items = dialogs.new_item(self.base_directory)
        else:
            if isinstance(new_rmd_file_name, RmdFile):
                new_rmd_file_name = new_rmd_file_name.full_path

            new_items = [RExamItem(RmdFile(new_rmd_file_name,
                                base_directory=self.base_directory)), None]
            if self.ig_l2.rexam_item is not None:
                new_items[1] = self.ig_l2.rexam_item
            elif self.ig_l1.rexam_item is not None:
                new_items[1] = self.ig_l1.rexam_item

        if new_items[0] is not None:
            self.save_items()
            fl_name = new_items[0].filename
            for n in new_items:
                if n is not None:
                    n.save()

            self.ig_l1.rexam_item = new_items[0]
            self.ig_l2.rexam_item = new_items[1] # FIXME no biligual new items
            if self.ig_l1.rexam_item.language_code == consts.CODE_L2:
                self.ig_l1.rexam_item, self.ig_l2.rexam_item = \
                    self.ig_l2.rexam_item, self.ig_l1.rexam_item # swap
            self.update_item_list()

            self.select_item_by_filename(fl_name)
            self.ig_l2.update_gui()
            self.ig_l1.update_gui()
            self.update_name()
            self.menu.update(menu_definition=self.menu_definition())

    def add_second_language(self):
            ifln = self.fl_list.files[self.idx_selected_item].rmdfile_l1
            fl_path = ifln.get_other_language_rmdfile()
            copy_content = sg.popup_yes_no("Copy content of {}?".format(
                        ifln.name))
            if copy_content == "Yes":
                new = ifln.copy_subdir_files(fl_path.name)
                if not isinstance(new, RmdFile):
                    # io error
                    log(new)
                    self.reset_gui()
                    return
                self.reset_gui()
            else:

                self.new_item(fl_path)

    def rename(self):
        n1, n2, fix_dir = dialogs.rename_item(self.lb_items.get()[0])
        if n1 is not None:
            self.save_items(ask=False)

            flns = self.fl_list.files[self.idx_selected_item]
            add_language = len(n2) and not flns.is_bilingual()
            # rename
            for new_name, old in zip((n1, n2),
                                     (flns.rmdfile_l1,
                                      flns.rmdfile_l2)):
                if new_name is not None and old is not None:
                    log(old.rename(new_name,
                                   new_sub_dir=new_name,
                                   rename_on_disk=True))
                    # TODO optional sub_dir renaming

            self.reset_gui()
            self.select_item_by_filename(n1 + RmdFile.SUFFIX)
            if add_language:
                self.add_second_language()
                self.reset_gui()
                self.select_item_by_filename(n1 + RmdFile.SUFFIX)



from os import listdir, path
import PySimpleGUI as sg

from .. import misc, consts
from ..rexam import extypes
from ..rexam.item import RExamItem, AnswerList
from ..rexam.git_info import GitInfo

_EMPTY_ITEM = RExamItem(None)

sg.theme_add_new("mytheme", consts.SG_COLOR_THEME)
sg.theme("mytheme")


class GUIBaseDirectory(object):

    def __init__(self, database_folder, label="Base Directory",
                 size=(60, 1), key="change_directory"):

        self.label = label
        self.text = sg.Text(database_folder, size=size,
                                background_color=consts.COLOR_BKG_ACTIVE_INFO)
        self.frame = sg.Frame(title=label, layout=[[self.text,
                      sg.Button("change", size=(6, 1), key=key)]], border_width=2)

    def update_folder(self, folder):
        self.text.update(value=folder)
        txt = self.label
        head = GitInfo(folder).head
        if len(head):
            txt += ": git head {}".format(head[:7])
        self.frame.update(value=txt)


def labeled_frame(elem, label="", border_width=0):
    if not isinstance(elem, list):
        elem = [elem]
    return sg.Frame(label,[elem], border_width=border_width)


def set_font(font, font_size):
    if font is not None:
        if font_size is None:
            font_size = ""
        sg.set_options(font='{} {}'.format(font, font_size))

class GUIItem(object):

    def __init__(self, label, key_prefix,
                 show_hash = True,
                 change_meta_info_button=False,
                 disabled = False):
        # all events start with key_prefix

        if consts.TAB_LAYOUT:
            len_ml = consts.LEN_ML_LARGE
            len_answer = consts.LEN_ANSWER_LARGE
        else:
            len_ml = consts.LEN_ML_SMALL
            len_answer = consts.LEN_ANSWER_SMALL

        self.disabled = disabled
        self.label = label
        self.show_hash = show_hash
        self.key_prefix = key_prefix
        self._item = None

        self.ml_quest = sg.Multiline(default_text="",
                                     size=(consts.WIDTH_ML, len_ml),
                                     enable_events=True,
                                     key="{}_quest".format(key_prefix))
        self.ml_answer = sg.Multiline(default_text="", enable_events=True,
                                      size=(consts.WIDTH_ML, len_answer),
                                      key="{}_answer".format(key_prefix))

        self.txt_answer_list = sg.Text("Answer-list", size=(10, 1),
                                       background_color=consts.COLOR_QUEST)

        self.ml_solution = sg.Multiline(default_text="", enable_events=True,
                                        size=(consts.WIDTH_ML, len_ml),
                                        key="{}_solution".format(key_prefix))
        self.ml_solution_answ_lst = sg.Multiline(default_text="", enable_events=True,
                                                 size=(
                                                     consts.WIDTH_ML, len_answer),
                                                 key="{}_solution_feedback".format(key_prefix))
        self.txt_solution_answ_lst = sg.Text("Answer-list", size=(10, 1),
                                             background_color=consts.COLOR_SOLUTION)

        self.ml_metainfo = sg.Multiline(default_text="",
                                        size=(consts.WIDTH_ML, 10), enable_events=True,
                                        key="{}_meta".format(key_prefix))

        self.change_meta_info_extra = sg.Button("Edit Meta Information",
                                                enable_events=True,
                                                visible=change_meta_info_button,
                                                key="{}_btn_change_meta".format(key_prefix))

        self.ml_info_validation =sg.Multiline(default_text="",
                                              size=(consts.WIDTH_ML - 26, 4),
                                              background_color=consts.COLOR_BKG_INACTIVE,
                                              disabled=True)

        self.ml_files = sg.Multiline(default_text="",
                                     size=(20, 4),
                                     background_color=consts.COLOR_BKG_INACTIVE,
                                     disabled=True,
                                     write_only=False)

        self.dd_types = sg.DropDown(values=[extypes.UNKNOWN_TYPE] +
                                           list(extypes.EXTYPES.keys()),
                                    size=(10,1),  enable_events=True,
                                    key="{}_dd_types".format(key_prefix))

        self.btn_add_answer_list = sg.Button("+", enable_events=True,
                                             size=(2, 1),
                                key="{}_btn_add_answer_list".format(key_prefix))
        self.btn_add_feedback_list = sg.Button("+", enable_events=True,
                                            size=(2,1),
                    key="{}_btn_add_feedback_list".format(key_prefix))

        self.btn_update_exsolution = sg.Button("update 'exsolution'",
                                               enable_events=True,
                                               button_color=consts.COLOR_RED_BTN,
                                               size=(15,1),
                                               key="{}_btn_update_exsolution".format(key_prefix))

        self.btn_fix_meta_issues = sg.Button("Auto-fix issues",
                                             enable_events=True,
                                             button_color=consts.COLOR_RED_BTN,
                                             size=(15,1),
                                             key="{}_btn_fix_meta_issues".format(key_prefix))

        # make main frame
        layout_question =[[self.ml_quest],
                        [self.txt_answer_list, self.btn_add_answer_list,
                         self.btn_update_exsolution],
                        [self.ml_answer]]

        layout_solution = [[self.ml_solution],
                        [self.txt_solution_answ_lst, self.btn_add_feedback_list],
                        [self.ml_solution_answ_lst]]

        layout_meta_info =  [[self.ml_metainfo],
                             [self.dd_types, self.btn_fix_meta_issues,
                              self.change_meta_info_extra]]

        if consts.TAB_LAYOUT:
            tab_group = sg.TabGroup([[sg.Tab("Question", layout_question,
                                             background_color=consts.COLOR_QUEST,
                                             key="{}_tab_quest".format(
                                                 self.key_prefix)),
                                      sg.Tab("Solution", layout_solution,
                                             background_color=consts.COLOR_SOLUTION,
                                             key="{}_tab_sol".format(
                                                 self.key_prefix))
                                      ]])
            self.gui_element = sg.Frame(self.label, [
                   [tab_group],
                   [sg.Frame("Meta-Information", layout_meta_info)],
                   [sg.Frame("Validation", [[self.ml_info_validation]]),
                    sg.Frame("Files", [[self.ml_files]])]
            ])
        else:
            self.gui_element = sg.Frame(self.label, [
                   [sg.Frame("Question", layout_question,
                             background_color=consts.COLOR_QUEST)],
                   [sg.Frame("Solution (feedback)", layout_solution,
                             background_color=consts.COLOR_SOLUTION)],
                    [sg.Frame("Meta-Information", layout_meta_info)],
                    [sg.Frame("", [[self.ml_info_validation]])]
            ])

    @property
    def rexam_item(self):
        return self._item

    @rexam_item.setter
    def rexam_item(self, v):
        assert isinstance(v, (RExamItem)) or v is None
        if self.disabled:
            self._item = None
        else:
            self._item = v
            self._activate_gui(v is not None)

    def is_activated(self):
        return self._item is not None

    def _activate_gui(self, value):
        if self.disabled:
            return
        if value:
            col = consts.COLOR_BKG_ACTIVE
        else:
            col = consts.COLOR_BKG_INACTIVE

        self.ml_quest.update(disabled=not value, background_color=col)
        self.ml_answer.update(disabled=not value, background_color=col)
        self.ml_solution.update(disabled=not value, background_color=col)
        self.ml_solution_answ_lst.update(disabled=not value,
                                         background_color=col)
        self.ml_metainfo.update(disabled=not value, background_color=col)
        self.dd_types.update(disabled=not value)
        self.change_meta_info_extra.update(disabled=not value)
        if not value:
            self.btn_add_answer_list.update(visible=False)
            self.btn_update_exsolution.update(visible=False)
            self.btn_add_feedback_list.update(visible=False)
            self.btn_fix_meta_issues.update(visible=False)

    def set_enable_answer_list(self, enable):
        if self.disabled:
            return

        if enable:
            col =  consts.COLOR_BKG_ACTIVE
        else:
            col = consts.COLOR_BKG_INACTIVE
        self.ml_answer.update(disabled=not enable, background_color=col)

        if self.is_activated():
            self.btn_add_answer_list.update(visible=not enable)

    def set_enable_feedback_list(self, enable):
        if self.disabled:
            return
        if enable:
            col =  consts.COLOR_BKG_ACTIVE
        else:
            col = consts.COLOR_BKG_INACTIVE
        self.ml_solution_answ_lst.update(disabled=not enable,
                                         background_color=col)
        if self.is_activated():
            self.btn_add_feedback_list.update(visible=not enable)

    def set_issues(self, issues):
        if self.disabled:
            return
        txt = ""
        auto_fix = False
        for i in issues:
            txt += "* {}\n".format(i.description)
            if i.has_fix_function:
                auto_fix = True

        self.ml_info_validation(value=txt)
        self.btn_fix_meta_issues.update(visible=auto_fix)
        if len(issues):
            self.ml_info_validation(background_color="darkred",
                                    text_color="white")
        else:
            self.ml_info_validation(background_color=consts.COLOR_BKG_ACTIVE_INFO,
                                    text_color="black")

    def as_markdown_file(self):
        if self.disabled:
            return ""

        rtn = "".join(self.rexam_item.header)
        rtn += _EMPTY_ITEM.question.str_markdown_heading()
        rtn += self.ml_quest.get().strip() + "\n\n"

        if len(self.ml_answer.get().strip())>0:
            rtn += AnswerList(_EMPTY_ITEM).str_markdown_heading()
            rtn += self.ml_answer.get().strip() + "\n\n"

        rtn += _EMPTY_ITEM.solution.str_markdown_heading()
        rtn += self.ml_solution.get().strip() + "\n\n"
        if len(self.ml_solution_answ_lst.get().strip())>0:
            rtn += AnswerList(_EMPTY_ITEM).str_markdown_heading()
            rtn += self.ml_solution_answ_lst.get().strip() + "\n\n"

        rtn += _EMPTY_ITEM.meta_info.str_markdown_heading()
        rtn += self.ml_metainfo.get().strip() + "\n"
        return rtn

    def update_answer_list_button(self):
        # extract solution and switch visibility
        if self.disabled:
            return

        if self.rexam_item is None or not self.rexam_item.question.has_answer_list_section():
            self.btn_update_exsolution.update(visible=False)
            return

        solution = AnswerList.extract_solution(self.ml_answer.get())
        self.btn_update_exsolution.update(visible=
                                          self.rexam_item.meta_info.solution != solution)

    def update_item(self):
        # new content from gui to ss_item
        self.rexam_item.parse(self.as_markdown_file(),
                              reset_meta_information=True)

    def update_gui(self):
        # copy items --> GUI elements and set active

        if self.disabled:
            return

        if self.rexam_item is None:
            item = _EMPTY_ITEM
        else:
            item = self.rexam_item


        fl_info = self.label + ": "
        tmp = path.join(item.sub_directory, item.filename)
        if len(tmp):
            if self.show_hash:
                fl_info += "[{}] ".format(item.hash_short())
            fl_info += "..." + path.sep + tmp
        self.gui_element.update(value=fl_info)

        self.ml_quest.update(value=item.question.str_text())
        self.ml_solution.update(value=item.solution.str_text())

        self.ml_metainfo.update(value=item.meta_info.str_parameter() +
                                item.meta_info.str_text())

        if not item.meta_info.check_type():
            t = extypes.UNKNOWN_TYPE
        else:
            t = item.meta_info.type
        self.dd_types.update(value=t)

        if not item.question.has_answer_list_section():
            self.ml_answer.update(value="")
        else:
            self.ml_answer.update(value=
                                  item.question.answer_list.str_answers(tag_mark_correct=True) +
                                  item.question.answer_list.str_text())


        if not item.solution.has_answer_list_section():
            self.ml_solution_answ_lst.update(value="")
        else:
            self.ml_solution_answ_lst.update(value=
                        item.solution.answer_list.str_answers() +
                        item.solution.answer_list.str_text())
        self.set_enable_answer_list(item.question.has_answer_list_section())
        self.set_enable_feedback_list(item.solution.has_answer_list_section())
        self.update_answer_list_button()

        # validation and file info
        if self.is_activated():
            col = consts.COLOR_BKG_ACTIVE_INFO
            self.set_issues(item.validate())
        else:
            col = consts.COLOR_BKG_INACTIVE
            self.ml_info_validation(value="", background_color=col)

        #files
        try:
            is_in_subfolder = len(self.rexam_item.sub_directory)>0
        except:
            is_in_subfolder = False

        if self.is_activated() and is_in_subfolder:
            self.ml_files.update(background_color=consts.COLOR_BKG_ACTIVE_INFO)
            x = misc.CaseInsensitiveStringList(listdir(item.directory))
            x.remove_all(item.filename)
            if len(x):
                self.ml_files(value="\n".join(x.get()))
            else:
                self.ml_files(value="")

        else:
            self.ml_files.update(background_color=consts.COLOR_BKG_INACTIVE)
            self.ml_files(value="")

    def save_item(self):
        if self.rexam_item is not None:
            self.update_item()
            self.rexam_item.save()
            self.update_gui()



import os
from os import path
from .. import misc
from .rmd_file import RmdFile, CODE_L1, CODE_L2, TAG_L1, TAG_L2, TAG_BILINGUAL
from .git_info import GitInfo

class BiLingualRmdFilePair(object):
    """Two RMD Files: l1 and l2
    """

    def __init__(self, rmd_file_a, rmd_file_b=None,
                 reference_language_code=None):
        """if reference language_code is defined, items that does not have
        this reference code will be stored as rmd_l2 """

        x = []
        for rmd in (rmd_file_a, rmd_file_b):
            if rmd is None:
                x.append(None)
            else:
                if isinstance(rmd, RmdFile):
                    x.append(rmd)
                else:
                    x.append(RmdFile(rmd))

        if x[0] is None and x[1] is not None:
            x = x[1], x[0]

        elif x[1] is not None:
            # --> two languages: swap, if x[1] is NL
            if reference_language_code is not None and \
                    x[1].language_code == reference_language_code:
                x = x[1], x[0]

        self._l1 = x[0]
        self._l2 = x[1]

    def __str__(self):
        if self._l1 is None:
            i = "None"
        else:
            i = self._l1.relative_path
        if self._l2 is None:
            t = "None"
        else:
            t = self._l2.relative_path

        return "{}: ({}, {})".format(self.shared_name(), i, t)

    @property
    def rmdfile_l1(self):
        return self._l1

    @property
    def rmdfile_l2(self):
        return self._l2

    def shared_name(self, add_bilingual_tag=True):

        if self._l1 is not None and self._l2 is None:
            return self._l1.name.lower()
        elif self._l1 is None and self._l2 is not None:
            return self._l2.name.lower()
        elif self._l1 is None and self._l2 is None:
            return None
        else:
            # is bilingual
            name = self._l1.name.lower()
            if name.endswith(TAG_L1) or \
                    name.endswith(TAG_L2):
                name = name[:-3]
            if add_bilingual_tag:
                name = name + TAG_BILINGUAL
            return name

    def is_bilingual(self):
        return self._l2 is not None and self._l1 is not None


class BiLingRmdFileList(object):

    def __init__(self, base_directory=None, files_first_level=True,
                                    files_second_level=True,
                                    check_for_bilingual_files=True):
        self.files = []
        self.base_directory = base_directory
        self.files_first_level = files_first_level
        self.files_second_level = files_second_level
        self.check_for_bilingual_files = check_for_bilingual_files

        # check for matching languages
        lst = misc.CaseInsensitiveStringList(self.get_rmd_files())

        self._file_list_hash = hash(tuple(lst.get())) # simple hashes for online
        # change detection, this are not the version IDs!

        if check_for_bilingual_files:
            reference_language = CODE_L1
        else:
            reference_language = None
        second = None
        while len(lst) > 0:
            first = RmdFile(file_path=lst.pop(0),
                            base_directory=self.base_directory)

            if check_for_bilingual_files:
                second = first.get_other_language_rmdfile()

            if second is not None and second.full_path in lst:
                # get in correct cases
                second_full_cases = lst.remove(second.full_path)
                lst.remove_all(second_full_cases) # remove all others versions (should not be the case)
                second = RmdFile(file_path=second_full_cases,
                                 base_directory=self.base_directory)
            else:
                second = None

            self.files.append(BiLingualRmdFilePair(rmd_file_a=first,
                                                   rmd_file_b=second,
                                                   reference_language_code=reference_language))

        self.files = sorted(self.files, key=lambda x:x.shared_name())

    def get_current_git_head_basedir(self):
        """returns the current git head of the based directory"""
        if self.base_directory is None:
            return ""
        else:
            return GitInfo(self.base_directory).head


    def get_count(self):

        rtn = {"total": len(self.files),
                CODE_L1: 0,
                CODE_L2: 0,
                "bilingual": 0,
                "undef": 0}

        for f in self.files:
            if f.is_bilingual():
                rtn["bilingual"] += 1
            else:
                try:
                    lang = f.rmdfile_l1.language_code
                except:
                    lang = None
                try:
                    rtn[lang] += 1
                except:
                    rtn["undef"] += 1

        return rtn

    def is_file_list_changed(self):
        lst = self.get_rmd_files()
        return hash(tuple(lst)) != self._file_list_hash

    def get_shared_names(self, bilingual_tag=True):
        return [x.shared_name(bilingual_tag) for x in self.files]

    def find_shared_name(self, name):
        tmp = self.get_shared_names(bilingual_tag=False)
        try:
            return tmp.index(name)
        except:
            return None

    def find_filename(self, fl_name):
        """find idx by file name of first or second language"""

        for cnt, fl in enumerate(self.files):
            if fl.rmdfile_l1.filename == fl_name or \
                    (fl.rmdfile_l2 is not None and
                     fl.rmdfile_l2.filename == fl_name):
                return cnt

        return None

    def get_rmd_files(self, suffix=RmdFile.SUFFIX):
        """returns list with Rmd files at the second levels and/or first
        level (depending on files_first_level and files_second_level).

         Second level files: The file with the same name as the folder;
         otherwise first rexam found is this subfolder return.

         returns absolute paths """

        if self.base_directory is None:
            return []
        lst = []
        for name in os.listdir(self.base_directory):
            fld = path.join(self.base_directory, name)
            if path.isdir(fld) and self.files_second_level:
                good_fl_name = path.join(fld, name + suffix)
                if path.isfile(good_fl_name):
                    lst.append(good_fl_name)
                else:
                    # search for a rexam file in the subdirectory
                    try:
                        subdir_lst = os.listdir(fld)
                    except:
                        subdir_lst = []  # no permission to access dir
                    for fl_name in map(lambda x: path.join(fld, x), subdir_lst):
                        # no permission
                        if fl_name.lower().endswith(suffix.lower()):
                            # best guess
                            lst.append(fl_name)
                            break

            elif path.isfile(fld) and self.files_first_level:
                if fld.lower().endswith(suffix.lower()):
                    lst.append(fld)

        return [path.abspath(x) for x in lst]





from os import path
from copy import deepcopy
import shutil
from .filepath import FilePath, os_rename
from ..consts import CODE_L1, CODE_L2

SEP = "-"
TAG_L1 = SEP + CODE_L1
TAG_L2 = SEP + CODE_L2
TAG_BILINGUAL = "{}[{}/{}]".format(SEP, CODE_L1, CODE_L2)


def _copytree(source_folder, destination_folder):
    """copies a folder and return error if it occurs"""
    try:
        shutil.copytree(source_folder, destination_folder)
    except IOError as io_error:
        return io_error


class RmdFile(FilePath):

    SUFFIX = ".Rmd"

    @staticmethod
    def make_path(base_directory, name, add_subdir=True):
        if add_subdir:
            return path.join(base_directory, name, "{}{}".format(name,
                                            RmdFile.SUFFIX))
        else:
            return path.join(base_directory, "{}{}".format(name,
                                                           RmdFile.SUFFIX))

    @property
    def language_code(self):
        if len(self.name)>=4 and self.name[-3] == SEP:
            lang = self.name[-2:].lower()
            if lang in (CODE_L1, CODE_L2):
                return lang
        return ""

    @language_code.setter
    def language_code(self, v):
        assert(isinstance(v, str) and len(v)==2)
        if self.name[-3] == SEP:
            self.name = self.name[-2] + v
        else:
            self.name = self.name + SEP + v

    def subdir_mirrors_filename(self):
        return self.name == self.sub_directory

    def get_other_language_rmdfile(self):
        if len(self.language_code):
            name = self.name[:-2]
            if self.language_code == CODE_L1:
                name += CODE_L2
            else:
                name += CODE_L1

            rtn = deepcopy(self)
            rtn.name = name
            if len(self.sub_directory):
                rtn.sub_directory = name
            return rtn
        else:
            return None

    def copy_subdir_files(self, new_name):
        """Returns io error, if it occurs other the new RmdFile object"""
        new = deepcopy(self)
        new.name = new_name
        new.sub_directory = new_name

        # copy files
        ioerror = _copytree(self.directory, new.directory)
        if ioerror:
            return ioerror
        else:
            copied_rmd = deepcopy(new)
            copied_rmd.name = self.name # has still old name
            ioerror = os_rename(copied_rmd.full_path, new.full_path)
            if ioerror:
                return ioerror

        return new



import os
import logging
import appdirs
from .. import APPNAME
from PySimpleGUI import Print

def get_log_file():
    appname = APPNAME.replace(" ", "_").lower()
    cache_dir = appdirs.user_cache_dir(appname)
    try:
        os.mkdir(cache_dir)
    except:
        pass
    return os.path.join(cache_dir, "{}.log".format(appname))


def init_logging():
        log_file = get_log_file()
        logging.basicConfig(level=logging.INFO,
                        format='[%(asctime)s]  %(message)s',
                        datefmt='%m-%d %H:%M:%S',
                        filename=log_file,
                        filemode='a')
        print("Log file: {}".format(log_file))

def log(txt, gui_log=True):

    if txt in (None, False):
        return

    for l in str(txt).split("\n"):
        if gui_log:
            Print(l)
            if l.startswith("WARNING"):
                Print("-" * 80 + "\n")
            elif l.startswith("ERROR"):
                Print("-" * 80 + "\n")

        if len(l)>0:
            if l.startswith("WARNING"):
                logging.warning(l)
            elif l.startswith("ERROR"):
                logging.error(l)
            else:
                logging.info(l)




import types

class Issue(object):
    # issues from the validations

    def __init__(self, label, description, fix_function=None,
                 fix_requires_gui_reset=False):
        self.label = label
        self.description = description
        self.fix_requires_gui_reset = fix_requires_gui_reset
        if isinstance(fix_function, (types.FunctionType, types.MethodType)):
            self.fix_fnc = fix_function
        else:
            self.fix_fnc = None

    def has_fix_function(self):
        return self.fix_fnc is not None

    def fix(self):
        if self.fix_fnc is not None:
            return self.fix_fnc()
        else:
            return False



from os import getcwd, path
from math import floor
import PySimpleGUI as sg
try:
    from markdown import markdown
except:
    markdown = None

from .. import __version__, APPNAME, consts
from .json_settings import JSONSettings
from .dialogs import ask_save
from ..rexam import exam
from .gui_misc import GUIBaseDirectory, labeled_frame, set_font


class ExamCompiler(object):
    SHOW_HASHES = True

    def __init__(self, settings=None, font=None, font_size=9):

        if not isinstance(settings, JSONSettings):
            self.settings = JSONSettings(
                         appname=APPNAME.replace(" ", "_").lower(),
                         settings_file_name="settings.json",
                         defaults= {"last_exam_file": None},
                         reset=False)
        else:
            self.settings = settings


        self.exam = exam.Exam()
        self.exam.item_database_folder = getcwd()

        set_font(font, font_size)
        self.gui_base_directory = GUIBaseDirectory(self.exam.item_database_folder)

        self.txt_exam = sg.Text("", size=(20, 1), background_color=consts.COLOR_BKG_ACTIVE_INFO)

        self.gui_db = GUIItemTable(show_l2=False,
                                   n_row=8,
                                   show_hash=False,
                                   short_hashes=True,
                                   key='database',
                                   tooltip='Item Database')
        self.gui_exam = GUIItemTable(show_l2=self.gui_db.show_l2,
                                     n_row=8,
                                     show_hash=False,
                                     short_hashes=self.gui_db.short_hashes,
                                     key='exam',
                                     tooltip='Exam Items')
        self.cb_language2 = sg.Checkbox('Second Language', key="cb_l2", enable_events=True)

        self.btn_save = sg.Button("Save", size=(12, 1), key="save_exam", disabled=True)

        self.layout = [
            [self.gui_base_directory.frame,
             labeled_frame([self.txt_exam, self.btn_save,
                            sg.Button("Load", size=(4, 1), key="load_exam"),
                            sg.InputText(visible=False, enable_events=True, key='new_exam'),
                            sg.FileSaveAs(button_text="New", file_types=(('json', '.json'),)),
                            ],
                           label="Exam", border_width=2),
             labeled_frame([self.cb_language2], label="Display", border_width=2),
             labeled_frame([sg.Button("Rexam code", size=(10, 1), key="btn_r_code"),
                            sg.Button("Close", size=(10, 1), key="btn_close")], label="Window", border_width=2),
             ],
            [labeled_frame([self.gui_db.table, self.gui_db.multiline], label="Database", border_width=2)],
            [
             sg.Button("add", size=(30, 2),
                       button_color= consts.COLOR_GREEN_BTN,
                       key="add_to_exam"),
            sg.Button("remove", size=(30, 2),
                      button_color=consts.COLOR_RED_BTN,
                      key="remove_from_exam"),
            sg.Button("up", size=(10, 2), key="move_up"),
            sg.Button("down", size=(10, 2), key="move_down")
            ],
            [sg.Frame("Exam", layout=[[self.gui_exam.table, self.gui_exam.multiline]])]
        ]

        self._selected_row_tab_db = None
        self.generate_html = False  # FIXME in GUI

    def _tmp_html_file(self):
        try:
            flname = self.exam.json_filename.replace(".json", "")
        except:
            return ""
        if len(flname):
            return path.abspath(flname) + ".tmp.html"
        else:
            return ""

    def update_tables(self, exam_tab_select_row=None):
        """table with item_id, name, short question l1 ,
           short question l2"""

        self.gui_base_directory.update_folder(self.exam.item_database_folder)
        # exam
        if self.exam.item_db is not None:
            db_ids = self.exam.get_database_ids(rm_nones=False)
        else:
            db_ids = []

        tmp = []
        for quest, idx in zip(self.exam.questions, db_ids):

            if idx is None:
                tmp.append(exam.EntryNotFound(quest, use_l2=True))
            else:
                tmp.append(self.exam.item_db.entries[idx])
        self.gui_exam.set_items(items=tmp)
        if exam_tab_select_row is not None:
            self.gui_exam.set_selected(exam_tab_select_row)

        # markdown and html
        md = self.exam.markdown(use_l2=self.gui_exam.show_l2)
        self.gui_exam.multiline.update(value=md)

        # save to .tmp.html file
        if self.generate_html and markdown is not None:
            flname = self._tmp_html_file()
            if len(flname):
                with open(flname, "w", encoding=consts.FILE_ENCODING) as fl:
                    fl.write(markdown(md))

        # not in exam --> show in database
        if self.exam.item_db is not None:
            tmp = [x for x in self.exam.item_db.selected_entries \
                                if x.id not in db_ids]
        else:
            tmp = []
        self.gui_db.set_items(items=tmp)

    @property
    def selected_db_row(self):
        return self._selected_row_tab_db

    @selected_db_row.setter
    def selected_db_row(self, v):
        if v != self._selected_row_tab_db:
            self._selected_row_tab_db = v
            if v is not None:
                cnt_selected = self.gui_db.get_row(v)[0]
                tmp = self.exam.item_db.entries[cnt_selected]

                if self.gui_db.show_l2:
                    tmp = tmp.item_l2
                else:
                    tmp = tmp.item_l1

                txt = "File {}\nHash {}\n\n".format(tmp.full_path, tmp.hash())
                txt += str(tmp)
                self.gui_db.multiline.update(value=txt)

    def new_exam(self, new_file):
        self.save_exam(ask=True)
        fld = self.exam.item_database_folder
        self.exam = exam.Exam()
        self.exam.json_filename = new_file
        self.exam.item_database_folder = fld
        self.reset_gui()
        self.save_exam(ask=False)

    def load_exam(self):
        self.save_exam(ask=True)
        try:
            default_path = path.split(self.exam.json_filename)[0]
        except:
            default_path = getcwd()
        json_filename = sg.PopupGetFile("", default_path = default_path,
                                        no_window = True,
                                        file_types = (('*.json', '*.json'),('ALL Files', '*.*')))
        try:
            self.exam.load(json_filename)
        except Exception as e:
            return e

        self.reset_gui()
        return True

    def save_exam(self, ask=True):
        if self.btn_save.Disabled:
            return

        try:
            already_saved_dict = exam.Exam(self.exam.json_filename).as_dict_list()
        except:
            already_saved_dict = None

        if already_saved_dict is None or already_saved_dict != self.exam.as_dict_list():
            # changes
            if ask: # something has changed
                if not ask_save(item_name=self.exam.json_filename):
                    return
            self.exam.save(self.exam.json_filename)

    def reset_gui(self):
        if self.exam.json_filename is None:
            name = ""
        else:
            name = path.split(self.exam.json_filename)[1]
        self.txt_exam.update(value=name)
        self.btn_save.update(disabled=not bool(len(name)))
        self.update_tables()

    def run(self, database_folder=None):

        win = sg.Window("{} ({})".format("Exam Compiler", __version__),
                        self.layout, finalize=True,
                        return_keyboard_events=True,
                        enable_close_attempted_event=True)

        if isinstance(database_folder, str):
            self.exam.item_database_folder = database_folder
        else:
            if self.settings.last_exam_file:
                try:
                    self.exam.load(self.settings.last_exam_file)
                except:
                    pass

        self.reset_gui()

        while True:
            win.refresh()
            event, values = win.read(timeout=5000)
            #print(event)

            if event == sg.WINDOW_CLOSE_ATTEMPTED_EVENT or \
                    event == "Close" or event == "btn_close":
                self.save_exam(ask=True)
                break

            elif event == "change_directory":
                fld = sg.PopupGetFolder(message="", no_window=True)
                if len(fld):
                    self.exam.item_database_folder = fld
                self.reset_gui()

            elif event=="save_exam":
                self.save_exam(ask=False)

            elif event=="load_exam":
                self.load_exam()

            elif event=="new_exam":
                self.new_exam(values[event])

            elif event=="btn_r_code":
                sg.Print(self.exam.rexam_code(use_l2=self.gui_exam.show_l2), size=(120, 40))

            elif event==self.gui_db.key_tab:

                try:
                    self.selected_db_row = values[event][0]
                except:
                    old = self.selected_db_row
                    self.selected_db_row = None
                    self.gui_db.set_selected(old)

            #elif event==self.gui_exam.key_tab:
            #    selected_entry = self.gui_exam.get_row()
            #    #TODO

            elif event == "cb_l2":
                self.gui_db.show_l2 = values[event]
                self.gui_exam.show_l2 = values[event]
                self.update_tables()

            elif event=="add_to_exam":
                try:
                    selected_entry = self.gui_db.get_row(values[self.gui_db.key_tab][0])
                except:
                    continue # nothing selected
                self.add_to_exam(selected_entry[0])

            elif event=="remove_from_exam":
                try:
                    selected_entry = self.gui_exam.get_row(values[self.gui_exam.key_tab][0])
                except:
                    continue # nothing selected
                self.remove_from_exam(selected_entry[0])

            elif event=="move_up":
                try:
                    selected_entry = values[self.gui_exam.key_tab][0]
                except:
                    continue # nothing selected
                self.exam.replace(selected_entry, selected_entry-1)
                self.update_tables(exam_tab_select_row=selected_entry - 1)

            elif event=="move_down":
                try:
                    selected_entry = values[self.gui_exam.key_tab][0]
                except:
                    continue # nothing selected
                self.exam.replace(selected_entry, selected_entry+1)
                self.update_tables(exam_tab_select_row=selected_entry + 1)

            else:
                pass#   print(event)

        win.close()

        self.settings.last_exam_file = self.exam.json_filename
        self.settings.save()

    def add_to_exam(self, selected_entry):
        item = self.exam.item_db.entries[selected_entry]
        self.exam.add_database_item(item)
        self.update_tables()

    def remove_from_exam(self, selected_entry):
        item = self.exam.item_db.entries[selected_entry]
        self.exam.remove_item(item)
        self.update_tables()


class GUIItemTable(object):
    LANGUAGES = ("Dutch", "English")

    def __init__(self, n_row, key, tooltip, max_lines = 1,
                 show_l2 = False, show_hash=True, short_hashes=True):
        self.max_lines = max_lines
        self.show_hash = show_hash
        self.short_hashes = short_hashes
        self.show_l2 = show_l2
        headings, width = self.get_headings()
        row_height = 30
        self.key_tab = key+"_tab"

        self.table = sg.Table(values=[[""] * len(headings)],
                              col_widths=width,
                              headings=[str(x) for x in range(len(headings))],
                              max_col_width=500,
                              background_color='white',
                              auto_size_columns=False,
                              display_row_numbers=False,
                              select_mode=sg.TABLE_SELECT_MODE_BROWSE,
                              justification='left',
                              num_rows=n_row,
                              enable_events=True,
                              bind_return_key=True,
                              # alternating_row_color='lightyellow',
                              key=self.key_tab,
                              row_height=row_height,
                              vertical_scroll_only=False,
                              tooltip=tooltip)

        h = floor(row_height * n_row/(1.4*10))
        self.multiline = sg.Multiline(size=(80, h))


    def update_headings(self):
        w = self.table.Widget
        if w is not None:
            for idx, txt in zip(self.table.ColumnHeadings, self.get_headings()[0]):
                w.heading(idx, text=txt)

    def get_headings(self):
        headings = ["Cnt", "Name",
                    GUIItemTable.LANGUAGES[int(self.show_l2)]]
        width = [5, 10, 70]
        if self.show_hash:
            headings.append("Hash")
            width.append(10)
        return headings, width

    def get_row(self, row_num):
        return self.table.get()[row_num]

    def set_items(self, items):
        values = []
        for x in items:
            d = [x.id]
            d.extend(x.short_repr(self.max_lines,
                                  use_l2=self.show_l2,
                                  add_versions=self.show_hash,
                                  short_version=self.short_hashes)) # TODO short hashes
            values.append(d)
        self.table.update(values=values)
        self.update_headings()

    def set_selected(self, selected):
        if not isinstance(selected, (list, tuple)):
            selected = [selected]
        self.table.update(select_rows=selected)



import os
import json
import time

from ..consts import FILE_ENCODING
from .item_database import ItemDatabase, BiLingualRmdFilePair, EntryItemDatabase
from .item import RmdFile, RExamItem


def _get_relpath_and_hash(item):
    # helper function
    try:
        return item.relative_path, \
               item.hash()
    except:
        return None, None


class ExamQuestion:

    def __init__(self, shared_name, path_l1, path_l2,
                 hash_l1, hash_l2):

        self.shared_name = shared_name
        self.path_l1 = path_l1
        self.path_l2 = path_l2
        self.hash_l1 = hash_l1
        self.hash_l2 = hash_l2

    def is_same_as(self, other):
        return isinstance(other, ExamQuestion) and \
               self.shared_name == other.shared_name and \
               self.path_l1 == other.path_l1 and \
               self.path_l2 == other.path_l2 and \
               self.hash_l1 == other.hash_l1 and \
               self.hash_l2 == other.hash_l2

    @staticmethod
    def create_from_database_entry(db_item):
        assert isinstance(db_item, EntryItemDatabase)
        ip, ih = _get_relpath_and_hash(db_item.item_l1)
        tp, th = _get_relpath_and_hash(db_item.item_l2)

        return ExamQuestion(shared_name=db_item.shared_name,
                            path_l1=ip, path_l2=tp,
                            hash_l1=ih, hash_l2=th)

    def markdown_l1(self):
        return RExamItem(RmdFile(self.path_l1)).markdown()

    def markdown_l2(self):
        return RExamItem(RmdFile(self.path_l2)).markdown()


class Exam(object):

    def __init__(self, json_filename=None):
        """
        """
        self.questions = []
        self._time_last_change = None
        self._item_db_folder = None
        self.info = None
        self.json_filename = None
        self.item_db = None
        self.git_head = ""
        if json_filename is not None:
            self.load(json_filename)

    @property
    def item_database_folder(self):
        return self._item_db_folder

    @item_database_folder.setter
    def item_database_folder(self, v):
        if isinstance(v, str) and os.path.isdir(v):
            self._item_db_folder = v
            self.item_db = ItemDatabase(v,
                                        files_first_level=True,
                                        files_second_level=True,
                                        check_for_bilingual_files=True)
        else:
            self._item_db_folder = None
            self.item_db = None

    @staticmethod
    def time_stamp():
        return time.strftime("%d-%m-%Y %H:%M", time.gmtime())

    def clear(self):
        self.questions = []
        self._time_last_change = Exam.time_stamp()

    def add_database_item(self, item):
        assert isinstance(item, (EntryItemDatabase, BiLingualRmdFilePair))
        if isinstance(item, BiLingualRmdFilePair):
            item = EntryItemDatabase.load(item, shared_name_with_bilingual_tag=False)

        self._time_last_change = Exam.time_stamp()
        path_l1, hash_l1 = _get_relpath_and_hash(item.item_l1)
        path_l2, hash_l2 = _get_relpath_and_hash(item.item_l2)
        self.questions.append(ExamQuestion(shared_name=item.shared_name,
                                           path_l1=path_l1,
                                           path_l2=path_l2,
                                           hash_l1=hash_l1,
                                           hash_l2=hash_l2))

    def reset_git_head(self):
        if isinstance(self.item_db, ItemDatabase):
            self.git_head = self.item_db.get_current_git_head_basedir()
        else:
            self.git_head = ""

    def is_incorrect_git_head(self): # TODO GUI needs to check this after load
        if len(self.git_head) and isinstance(self.item_db, ItemDatabase):
            return self.git_head != self.item_db.get_current_git_head_basedir()
        else:
            return False

    def as_dict_list(self):
        return {"time" : self._time_last_change,
                "item_database_folder": self.item_database_folder,
                "git_head": self.git_head,
                "names": [x.shared_name for x in self.questions],
                "paths_l1": [x.path_l1 for x in self.questions],
                "paths_l2" : [x.path_l2 for x in self.questions],
                "hashes_l1": [x.hash_l1 for x in self.questions],
                "hashes_l2": [x.hash_l2 for x in self.questions]}

    def save(self, json_filename=None, info=None, reset_git_head=True):
        if json_filename is None:
            if self.json_filename is None:
                raise RuntimeError("Specify json_filename to save exam.")
            else:
                json_filename = self.json_filename

        if reset_git_head:
            self.reset_git_head()

        print("Save {}".format(json_filename))
        if info is not None:
            self.info = info

        d = self.as_dict_list()
        if self.info is not None:
            d["info"] = self.info

        with open(json_filename, 'w', encoding=FILE_ENCODING) as fl:
            fl.write(json.dumps(d, indent = 2))

    def load(self, json_filename):

        with open(json_filename, 'r', encoding=FILE_ENCODING) as fl:
            d = json.load(fl)

        try:
            self._time_last_change = d["time"]
        except:
            self._time_last_change = None
        try:
            self.info = d["info"]
        except:
            self.info = None
        try:
            self.item_database_folder = d["item_database_folder"]
        except:
            self.item_database_folder = None
        try:
            self.git_head = d["git_head"]
        except:
            self.git_head = ""
        try:
            names = d["names"]
        except:
            names = []

        if self.is_incorrect_git_head():
            print("WARNING: incorrect git commit in database")

        self.questions = []
        for x in range(len(names)):
            self.questions.append(ExamQuestion(shared_name=d["names"][x],
                                               path_l1=d["paths_l1"][x],
                                               path_l2=d["paths_l2"][x],
                                               hash_l1=d["hashes_l1"][x],
                                               hash_l2=d["hashes_l2"][x]))
        self.json_filename = json_filename

    def get_database_ids(self, rm_nones=False):
        """returns ids from item database or the question if not found
        takes into account the hashes!"""

        if self.item_db is None:
            return []

        rtn = []
        for quest in self.questions:
            idx = self.item_db.find(
                hash_l1=quest.hash_l1 ,
                hash_l2=quest.hash_l2,
                shared_name=quest.shared_name,
                relative_path_l1=quest.path_l1,
                relative_path_l2=quest.path_l2,
                find_all=False) # finds just first one

            if idx is not None:
                rtn.append(idx)
            elif not rm_nones:
                rtn.append(None)

        return rtn

    def find_item(self, item):
        """returns question id first occurance if item"""

        if not isinstance(item, EntryItemDatabase):
            return None
        else:
            needle = ExamQuestion.create_from_database_entry(item)
            for x, quest in enumerate(self.questions):
                if quest.is_same_as(needle):
                    return x
            return None

    def remove_item(self, item):
        idx = self.find_item(item)
        if idx is not None:
            return self.questions.pop(idx)

    def replace(self, source_id, target_id):
        if source_id < len(self.questions) and target_id < len(self.questions):
            tmp = self.questions.pop(source_id)
            self.questions = self.questions[:target_id] + [tmp] + \
                             self.questions[target_id:]
            return True
        else:
            return False

    def markdown(self, use_l2=False, mark_correct=True):
        if self.item_db is None:
            return ""

        rtn = ""
        for cnt, db_idx in enumerate(self.get_database_ids(rm_nones=False)):
            if db_idx is not None:
                db_entry = self.item_db.entries[db_idx]
            else:
                db_entry = EntryNotFound(self.questions[cnt])

            if use_l2:
                tmp = db_entry.item_l2
            else:
                tmp = db_entry.item_l1

            if mark_correct:
                mark = "**"
            else:
                mark = ""
            q_str = tmp.markdown(enumerator=cnt + 1, wrap_text_width=80,
                                 tag_mark_correct=False,
                                 highlight_correct_char=mark)
            rtn += q_str + "\n\n"

        return rtn.strip()

    def rexam_code(self, use_l2=False):

        code = "library(exams)\n\n"
        code += "questions = c(\n"
        for quest, db_id in zip(self.questions, self.get_database_ids(rm_nones=False)):
            if use_l2:
                p = quest.path_l2
                h = quest.hash_l2
            else:
                p = quest.path_l1
                h = quest.hash_l1

            if db_id:
                code += " "*6 + "'{}',\n".format(p)
            else:
                code += "#" + " " *5 + "'{}',  # CAN NOT FIND ITEM {} \n".format(p, h)

        code = code.rstrip()[:-1] + "\n)\n"

        code += """
exams2pdf(questions,
  encoding = 'UTF-8',
  edir = {},
  verbose = TRUE
)
""".format(self.item_database_folder)

        return code



class EntryNotFound(EntryItemDatabase):

    def __init__(self, exam_question, use_l2=False):
        assert isinstance(exam_question, ExamQuestion)

        l = "-"*79 + "\n"
        tarray = [l, "ERROR: File not found\n", l]
        item_l1 = RExamItem()
        item_l2 = RExamItem()
        if not use_l2:
            item_l1.question.text_array = tarray +\
                                      ["File: {}\n".format(exam_question.path_l1),
                                       "Hash: {}\n".format(exam_question.hash_l1)]
            item_l1.name = exam_question.shared_name
            item_l1.meta_info.name = exam_question.shared_name
        else:
            item_l2.question.text_array = tarray +\
                                      ["File: {}\n".format(exam_question.path_l2),
                                       "Hash: {}\n".format(exam_question.hash_l2)]
            item_l2.name = exam_question.shared_name
            item_l2.meta_info.name = exam_question.shared_name

        super().__init__(shared_name=exam_question.shared_name,
                         item_l1=item_l1, item_l2=item_l2)



from os import path, makedirs, rename
from copy import deepcopy

def os_rename(source, destination):
    """rename file or folder and return error if it occurs"""
    try:
        return rename(source, destination)
    except IOError as io_error:
        return io_error

class FilePath(object):

    def __init__(self, file_path, base_directory = None):
        """Class handling filename and relative file locations

        Base_directory, sub_directory (relative to base) and filename

        if base directory is None,  base_directory assumed to be at first
        level and subdir at second level

        otherwise: file_path is relative to base directory
        """

        if file_path is None or file_path == "":
            self.filename = ""
            self.sub_directory = ""
            if base_directory is None:
                self.base_directory = ""
            else:
                self.base_directory = base_directory

        elif base_directory is None:
            directory, self.filename = path.split(path.abspath(file_path))
            self.base_directory, self.sub_directory = path.split(directory)
            if len(self.base_directory)==0:
                self.sub_directory, self.base_directory = \
                                    self.base_directory, self.sub_directory

        else:
            self.base_directory = path.abspath(base_directory)
            abs_file_path = path.abspath(file_path)
            if abs_file_path.startswith(self.base_directory):
                # abs_file_path is subpart of base directory
                rel_file_path = path.relpath(abs_file_path,
                                             start= self.base_directory)
            else:
                # abs_file_path is not subpart of base directory,
                # it's thus relative to base_directory
                rel_file_path = file_path

                if file_path == abs_file_path:
                    # it's abs_solute path (but not part of base_directory)
                    # base_direct does not make sense
                    self.base_directory = ""

            self.sub_directory, self.filename = path.split(rel_file_path)

    def __eq__(self, other):
        try:
            return self.full_path == other.full_path
        except:
            return False

    def __str__(self):
        return self.full_path

    @property
    def name(self):
        # without file extension
        return path.splitext(self.filename)[0]

    @name.setter
    def name(self, value):
        # changes name (and keeps extension)
        ext = path.splitext(self.filename)[1]
        self.filename = value + ext

    @property
    def full_path(self):
        return path.join(self.base_directory, self.sub_directory, self.filename)

    @property
    def directory(self):
        return path.join(self.base_directory, self.sub_directory)

    @property
    def relative_path(self):
        """path relative to base director"""
        return path.join(self.sub_directory, self.filename)

    def make_dirs(self):
        try:
            makedirs(self.directory)
        except:
            pass

    def rename(self, new_name, new_sub_dir = None, rename_on_disk=False):
        """Returns io error, if it occurs"""
        new = deepcopy(self)
        new.name = new_name
        if new_sub_dir is not None:
            new.sub_directory = new_sub_dir

        if rename_on_disk:
            if new_sub_dir is not None:
                io_error = os_rename(self.directory, new.directory)
                if io_error:
                    return "Can't rename directory: {}".format(io_error)

            io_error = os_rename(path.join(new.directory, self.filename),
                                 new.full_path)
            if io_error:
                return "Can't rename file: {}".format(io_error)

        self.filename = new.filename
        self.sub_directory = new.sub_directory



import sys as _sys

PYTHON_VERSION = "{0}.{1}.{2}".format(_sys.version_info[0],
                                      _sys.version_info[1],
                                      _sys.version_info[2])


def info():
    from PySimpleGUI import version as sq_version
    from . import WEBSITE
    from .rexam.r_render import RPY2INSTALLED
    from .gui.mainwin import MainWin
    from .misc import yesno

    settings_file = MainWin().settings.settings_file
    return ["Python {}".format(PYTHON_VERSION),
        "PySimpleGui {}".format(sq_version),
        "RPy2 installed: {}".format(yesno(RPY2INSTALLED)),
        "Settings: {}".format(settings_file),
        "Web: {}".format(WEBSITE)]


import argparse
from . import sysinfo

def cli(app_name):
    """returns dict with selected command line option"""

    parser = argparse.ArgumentParser(
        description=app_name,
        epilog="(c) O. Lindemann, 2021")

    parser.add_argument("-c", "--clear", action="store_true",
                        help="reset app and clear settings file",
                        default=False)

    parser.add_argument("-i", "--info",
                        action="store_true",
                        help="display system info",
                        default=False)

    parser.add_argument("--monolingual",
                        action="store_true",
                        help="switch to monolingual mode (default is "
                             "bilingual)",
                        default=False)

    parser.add_argument("-e", "--exam",
                        action="store_true",
                        help="run exam compiler",
                        default=False)

    parser.add_argument("-r", "--rcode", nargs='?', metavar="JSON-FILE", type=str,
                        help="convert exam file to r code",
                        default=False)

    parser.add_argument("-m", "--markdown", nargs='?', metavar="JSON-FILE", type=str,
                        help="convert exam file to simple markdown",
                        default=False)

    parser.add_argument("--l2", action="store_true",
                        help="use second language when converting exam",
                        default=False)

    opt = vars(parser.parse_args())

    if opt["info"]:
        print(app_name)
        print("\n".join(sysinfo.info()))

    elif None in (opt["rcode"], opt["markdown"]):
        parser.print_help()
        print("")
        print("Please specify exam file.")

    elif opt["rcode"]:
        from .rexam import exam
        a = exam.Exam(opt["rcode"])
        print(a.rexam_code(use_l2=opt["l2"]))

    elif opt["markdown"]:
        from .rexam import exam
        a = exam.Exam(opt["markdown"])
        print(a.markdown(use_l2=opt["l2"]))

    elif opt["exam"]:
        from .gui.exam_compiler import ExamCompiler
        ExamCompiler().run()

    else:
        return opt

    exit()



from . import __version__, APPNAME
from .cli import cli


def run():

    opt = cli("{} {}".format(APPNAME, __version__))

    from .gui.mainwin import MainWin
    MainWin(clear_settings=opt["clear"],
            monolingual=opt["monolingual"]).run()


if __name__ == "__main__":
    run()



from multiprocessing import Pool

from .rmd_file_list import BiLingRmdFileList, BiLingualRmdFilePair
from .item import RExamItem
from ..misc import iter_list


class _SearchSchemata(object):

    def __init__(self):
        self.search_types = []
        self.parameter = []
        self.functions = []

    def add(self, search_type, parameter):
        """para: search_type has to be 'name', 'question', 'solution',
        'meta_info' or 'raw_rmd'
        if parameter is iterable (but not a string), it will be multiple
        functions with different parameter.
        """

        if not isinstance(parameter, str) and hasattr(parameter, "__iter__"):
            for p in parameter:
                self._add(search_type, p)
        else:
            self._add(search_type, parameter)

    def _add(self, search_type, parameter):
        if search_type == "name":
            f = lambda x: x.name.find(parameter)
        elif search_type == "question":
            f = lambda x: x.question.str_text().find(parameter)
        elif search_type == "solution":
            f = lambda x: x.solution.str_text().find(parameter)
        elif search_type == "meta_info":
            f = lambda x: x.meta_info.str_text().find(parameter)
        elif search_type == "raw_rmd":
            f = lambda x: str(x).find(parameter)
        else:
            raise RuntimeError("{} is an unknown SearchFunction "
                               "type.".format(search_type))

        self.functions.append(f)
        self.parameter.append(parameter)
        self.search_types.append(search_type)


class EntryItemDatabase(object):
    # biligual RExamItem

    def __init__(self, shared_name, item_l1, item_l2):
        assert isinstance(item_l1, RExamItem) or item_l1 is None
        assert isinstance(item_l2, RExamItem) or item_l2 is None
        self.shared_name = shared_name
        self.item_l1 = item_l1
        self.item_l2 = item_l2
        self.id = None

    def is_same_as(self, item):
        """compares shared names and version id
        and ignores the id"""

        if isinstance(item, EntryItemDatabase):
            return self.shared_name == item.shared_name and \
                   self.hash_l1() == item.hash_l1() and \
                   self.hash_l2() == item.hash_l2()
        else:
            return False

    def hash_l1(self):
        try:
            return self.item_l1.hash()
        except:
            return ""

    def hash_l2(self):
        try:
            return self.item_l2.hash()
        except:
            return ""

    def hash_l1_short(self):
        try:
            return self.item_l1.hash_short()
        except:
            return ""

    def hash_l2_short(self):
        try:
            return self.item_l2.hash_short()
        except:
            return ""

    def short_repr(self, max_lines, use_l2, add_versions=False, short_version=True):
        if use_l2:
            try:
                txt = self.item_l2.question.str_text_short(max_lines, ignore_empty_lines=True)
            except:
                txt = ""
        else:
            try:
                txt = self.item_l1.question.str_text_short(max_lines, ignore_empty_lines=True)
            except:
                txt = ""

        rtn = [self.shared_name, txt]
        if add_versions:
            if short_version:
                if use_l2:
                    rtn.append(self.hash_l2_short())
                else:
                    rtn.append(self.hash_l1_short())
            else:
                if use_l2:
                    rtn.append(self.hash_l2())
                else:
                    rtn.append(self.hash_l1())
        return rtn

    @staticmethod
    def load(biling_filelist_entry, shared_name_with_bilingual_tag=False):
        assert isinstance(biling_filelist_entry, BiLingualRmdFilePair)

        if biling_filelist_entry.rmdfile_l1 is not None:
            l1 = RExamItem(biling_filelist_entry.rmdfile_l1)
        else:
            l1 = None

        if biling_filelist_entry.rmdfile_l2 is not None:
            l2 = RExamItem(biling_filelist_entry.rmdfile_l2)
        else:
            l2 = None

        return EntryItemDatabase(
                shared_name=biling_filelist_entry.shared_name(
                                    add_bilingual_tag=shared_name_with_bilingual_tag),
                item_l1=l1,
                item_l2=l2)


class ItemDatabase(BiLingRmdFileList):

    def __init__(self, base_directory, files_first_level,
                            files_second_level,
                            check_for_bilingual_files):
        """file_list_bilingual: path or file_list_biligual
        """
        super().__init__(base_directory=base_directory,
                         files_first_level=files_first_level,
                         files_second_level=files_second_level,
                         check_for_bilingual_files=check_for_bilingual_files)
        self._selected_ids = []

        ## LOAD DATA
        if len(self.files) > 1000:
            # speed up with multiprocessing
            entries = Pool().map(EntryItemDatabase.load, self.files)
        else:
            entries = map(EntryItemDatabase.load, self.files)
        self.entries = list(entries)

        # add unique ids
        for x in range(len(self.entries)):
            self.entries[x].id = x

        self.select() # select all

    def get_entries(self, ids, rm_nones=True):
        """returns subset of DatabaseEntries items """
        rtn = []
        for i in ids:
            try:
                rtn.append(self.entries[i])
            except:
                if not rm_nones:
                    rtn.append(None)
        return rtn

    @property
    def selected_entries(self):
        """selected name, l1, l2, l1_hash,
                    l2_hash"""
        return self.get_entries(self._selected_ids, rm_nones=True)

    def _search_select(self, search_function, item_ids_subset):
        """searches rexam file using search_function and returns idx,
        if found for at least one of the language.
        Use item_ids_subset (array if ids) to define the the  subset of
        items, in which you want to serach """

        idx = []
        for x in item_ids_subset:

            try:
                found = search_function(self.entries[x].item_l1)
            except:
                found = -1

            if found<0:
                try:
                    found = search_function(self.entries[x].item_l2)
                except:
                    found = -1

            if found>=0:
                idx.append(x)

        return idx

    def select(self, name=None, question=None, solution=None,
                     meta_information=None, raw_rmd=None,
                     search_logic_or=False):
        """select items based on text search"""

        # select all
        self._selected_ids = range(len(self.entries))
        search = _SearchSchemata()
        search.add("name", iter_list(name))
        search.add("question", iter_list(question))
        search.add("solution", iter_list(solution))
        search.add("meta_info", iter_list(meta_information))
        search.add("raw_rmd", iter_list(raw_rmd))

        if search_logic_or:
            # OR
            idx = []
            for fnc in search.functions:
                for x in self._search_select(fnc, item_ids_subset=self._selected_ids):
                    if x not in idx:
                        idx.append(x)
            self._selected_ids = sorted(idx)
        else:
            # AND
            for fnc in search.functions:
                self._selected_ids = self._search_select(fnc, item_ids_subset=self._selected_ids)

        return self._selected_ids

    def find_entry(self, entry_item_database):
        """returns all id of identical entries """

        same = filter(lambda x:x.is_same_as(entry_item_database), self.entries)
        rtn = map(lambda x: x.id, same)
        return list(rtn)

    def find(self, hash_l1,
             hash_l2=None,
             relative_path_l1 = None,
             relative_path_l2=None,
             shared_name = None,
             find_all=False):

        """advanced search function
        returns on first id  or array with all is (find_all=True)
        the found entries
        """

        rtn = []
        for cnt, e in enumerate(self.entries):
            a = shared_name is None or e.shared_name == shared_name
            b = relative_path_l1 is None or (e.item_l1 is not None and
                                             e.item_l1.relative_path == relative_path_l1)
            c = relative_path_l2 is None or \
                (e.item_l2 is not None and \
                 e.item_l2.relative_path == relative_path_l2)

            if  a and b and c:
                if  e.hash_l1() == hash_l1 and \
                    (hash_l2 is None or
                     e.hash_l2() == hash_l2):
                    if find_all:
                        rtn.append(cnt)
                    else:
                        return cnt

        if not find_all:
            return None
        else:
            return rtn



import os
from .. import APPNAME, misc

try:
    import rpy2.robjects as robjects
    import webbrowser
except:
    robjects = None

RPY2INSTALLED = robjects is not None


class RRender(object):

    HTML_FILE_NAME = "html-test"

    def __init__(self):
        self.html_dir = misc.get_temp_dir(
                    APPNAME.replace(" ","_").lower(), make_dir=True)

        r_code = '''library(exams)
        rmd_to_html <- function(filename) {
                exams2html(filename, quiet=TRUE, ''' + \
                 'dir="{}",'.format(self.html_dir) + \
                 'encoding = "utf8", template = "plain8", ' + \
                 'name="{}"'.format(RRender.HTML_FILE_NAME) + ")}\n"

        if robjects is not None:
            try:
                robjects.r(r_code)
                self.r_init_error = None
            except Exception as error:
                self.r_init_error = error
        else:
            self.r_init_error = "rpy2 is not installed."

    def rmd_to_html(self, file):
        """:return error text or None, if conversion succeeded"""
        if self.r_init_error is not None:
            return self.r_init_error
        r_func = robjects.r['rmd_to_html']
        try:
            r_func(file)
            return None
        except Exception as error:
            return error

    def get_html_file(self):
        return os.path.join(self.html_dir,
                            RRender.HTML_FILE_NAME + "1.html")

    def open_html(self, new=0):
        fl = self.get_html_file()
        if os.path.isfile(fl):
            webbrowser.open(fl, new=new)


from os import path
import yaml

class GitInfo(object):

    def __init__(self, folder):
        self._git = path.join(folder, ".git")
        try:
            with open(path.join(self._git, "HEAD"), "r") as fl:
                self._head = yaml.safe_load(fl)
        except:
            self._head = None

    @property
    def is_repository(self):
        return self._head is not None

    @property
    def folder(self):
        return self._git[:-4]


    @property
    def branch(self):
        if self._head is None:
            return ""

        elif isinstance(self._head, dict):
            return self._head['ref'].split(path.sep)[-1]
        else:
            return self._head.strip()

    @property
    def head(self):
        if self._head is None:
            return ""

        elif isinstance(self._head, dict):
            fl = path.join(self._git, self._head['ref'])
            with open(fl, "r") as fl:
                return fl.read().strip()
        else:
            return self._head.strip()



__author__  = "Oliver Lindemann"
__version__ = "0.5"

import os
import json
from copy import copy
from appdirs import user_config_dir



class JSONSettings():

    FILE_ENCODING = 'utf-8'

    def __init__(self, appname, settings_file_name, defaults, reset=False):
        """General settings class that saves the settings in a json file.
        If no config file exists it will be created.
        The variables (fields) can be access a properties of the Settings object.

        Changes DEFAULTS to defined the required settings and their defaults.

        if reset==True, old setting file will deleted.
        """

        self.config_dir = user_config_dir(appname=appname)
        try:
            os.makedirs(self.config_dir)
        except:
            pass
        self.settings_file = os.path.join(self.config_dir, settings_file_name)
        if reset:
            try:
                os.remove(self.settings_file)
            except:
                pass

        try: #load
            with open(self.settings_file, 'r',
                      encoding=JSONSettings.FILE_ENCODING) as\
                    fl:
                settings = json.load(fl)
            saving_required = False
        except:
            settings = defaults
            saving_required = True

        # check all defaults
        for k in defaults:
            if k not in settings:
                settings[k] = defaults[k]
                saving_required = True

        self._fields = settings.keys()
        self.__dict__.update(settings)

        if saving_required:
            self.save()

    def get_dict(self):
        return self.__dict__

    def save(self):
        d = copy(self.__dict__)
        try:
            d.pop("_fields")
        except:
            pass
        try:
            d.pop("settings_file")
        except:
            pass
        try:
            d.pop("config_dir")
        except:
            pass

        with open(self.settings_file, 'w',
                  encoding=JSONSettings.FILE_ENCODING) as fl:
            fl.write(json.dumps(d, indent = 2))



#!/usr/bin/env python3
"""
Installer
"""

from setuptools import setup, find_packages
import codecs
import os
from sys import version_info as _vi

from rexam_item_editor import __version__ as version
from rexam_item_editor import WEBSITE

package_name = "rexam_item_editor"
application_name = "rexam-item-editor"

install_requires = ["appdirs>=1.4",
                    "pysimplegui>=4.33"]

extras_require = {"rendering": ["rpy2>=3.4.3"]}

entry_points = {'console_scripts':
                ['{}={}.__main__:run'.format(application_name, package_name)]}

package_data = {'' : ['*.png', '*.Rmd']}


if _vi.major< 1:
    raise RuntimeError("{0} requires Python 3 or larger.".format(application_name))

def readme():
    directory = os.path.dirname(os.path.join(
        os.getcwd(), __file__, ))
    with codecs.open(
        os.path.join(directory, "README.md"),
        encoding="utf8",
        mode="r",
        errors="replace",
        ) as file:
        return file.read()

def get_version(package):
    """Get version number"""

    with open(os.path.join(package, "__init__.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("'")[1]
    return "None"


if __name__ == '__main__':
    setup(
        name = application_name,
        version=version,
        description='Maintaining and editing databases of RExam items',
        author='Oliver Lindemann',
        author_email='lindemann@cognitive-psychology.eu',
        license='MIT licence',
        url=WEBSITE,
        packages=find_packages(),
        include_package_data=True,
        package_data=package_data,
        setup_requires=[],
        install_requires=install_requires,
        entry_points=entry_points,
        extras_require=extras_require,
        keywords = "",
        classifiers=[
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering"
        ],
        long_description=readme(),
        long_description_content_type='text/markdown'
    )



UNKNOWN_TYPE = "unknown"
EXTYPES = {"schoice": "Multiple choice",
           "mchoice": "Multiple answer",
           "num": "Fill in the blank numbers",
           "string": "Fill in the blank text/essay",
           "cloze": "Combinations"}

HAVE_ANSWER_LIST = ["schoice", "mchoice", "cloze"]



from os  import path as _path

_fldir = _path.dirname(__file__)

FILES = {"schoice": _path.join(_fldir, "TemplateMultipleChoice.Rmd"),
        "mchoice": _path.join(_fldir, "TemplateMultipleAnswer.Rmd"),
        "num": _path.join(_fldir, "TemplateBlankNumber.Rmd"),
        "string": _path.join(_fldir, "TemplateBlankTextEssay.Rmd"),
        "cloze": _path.join(_fldir, "TemplateBlankCombination.Rmd")}